import sys
from typing import Any

import numpy as np
import numpy.typing as npt
from numpy._typing import _32Bit,_64Bit, _128Bit

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

# Can't directly import `np.float128` as it is not available on all platforms
f16: np.floating[_128Bit]

c16 = np.complex128()
f8 = np.float64()
i8 = np.int64()
u8 = np.uint64()

c8 = np.complex64()
f4 = np.float32()
i4 = np.int32()
u4 = np.uint32()

dt = np.datetime64(0, "D")
td = np.timedelta64(0, "D")

b_ = np.bool()

b = bool()
c = complex()
f = float()
i = int()

AR_b: npt.NDArray[np.bool]
AR_u: npt.NDArray[np.uint32]
AR_i: npt.NDArray[np.int64]
AR_f: npt.NDArray[np.float64]
AR_c: npt.NDArray[np.complex128]
AR_m: npt.NDArray[np.timedelta64]
AR_M: npt.NDArray[np.datetime64]
AR_O: npt.NDArray[np.object_]
AR_number: npt.NDArray[np.number[Any]]

AR_LIKE_b: list[bool]
AR_LIKE_u: list[np.uint32]
AR_LIKE_i: list[int]
AR_LIKE_f: list[float]
AR_LIKE_c: list[complex]
AR_LIKE_m: list[np.timedelta64]
AR_LIKE_M: list[np.datetime64]
AR_LIKE_O: list[np.object_]

# Array subtraction

assert_type(AR_number - AR_number, npt.NDArray[np.number[Any]])

assert_type(AR_b - AR_LIKE_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_b - AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_b - AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_b - AR_LIKE_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_b - AR_LIKE_m, npt.NDArray[np.timedelta64])
assert_type(AR_b - AR_LIKE_O, Any)

assert_type(AR_LIKE_u - AR_b, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_i - AR_b, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f - AR_b, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_c - AR_b, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_m - AR_b, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_M - AR_b, npt.NDArray[np.datetime64])
assert_type(AR_LIKE_O - AR_b, Any)

assert_type(AR_u - AR_LIKE_b, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_u - AR_LIKE_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_u - AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_u - AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_u - AR_LIKE_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_u - AR_LIKE_m, npt.NDArray[np.timedelta64])
assert_type(AR_u - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_u - AR_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_i - AR_u, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f - AR_u, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_c - AR_u, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_m - AR_u, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_M - AR_u, npt.NDArray[np.datetime64])
assert_type(AR_LIKE_O - AR_u, Any)

assert_type(AR_i - AR_LIKE_b, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i - AR_LIKE_u, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i - AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i - AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_i - AR_LIKE_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_i - AR_LIKE_m, npt.NDArray[np.timedelta64])
assert_type(AR_i - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_u - AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_i - AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f - AR_i, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_c - AR_i, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_m - AR_i, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_M - AR_i, npt.NDArray[np.datetime64])
assert_type(AR_LIKE_O - AR_i, Any)

assert_type(AR_f - AR_LIKE_b, npt.NDArray[np.floating[Any]])
assert_type(AR_f - AR_LIKE_u, npt.NDArray[np.floating[Any]])
assert_type(AR_f - AR_LIKE_i, npt.NDArray[np.floating[Any]])
assert_type(AR_f - AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_f - AR_LIKE_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_f - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_u - AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_i - AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_f - AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_c - AR_f, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_O - AR_f, Any)

assert_type(AR_c - AR_LIKE_b, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_c - AR_LIKE_u, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_c - AR_LIKE_i, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_c - AR_LIKE_f, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_c - AR_LIKE_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_c - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_u - AR_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_i - AR_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_f - AR_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_c - AR_c, npt.NDArray[np.complexfloating[Any, Any]])
assert_type(AR_LIKE_O - AR_c, Any)

assert_type(AR_m - AR_LIKE_b, npt.NDArray[np.timedelta64])
assert_type(AR_m - AR_LIKE_u, npt.NDArray[np.timedelta64])
assert_type(AR_m - AR_LIKE_i, npt.NDArray[np.timedelta64])
assert_type(AR_m - AR_LIKE_m, npt.NDArray[np.timedelta64])
assert_type(AR_m - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_m, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_u - AR_m, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_i - AR_m, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_m - AR_m, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_M - AR_m, npt.NDArray[np.datetime64])
assert_type(AR_LIKE_O - AR_m, Any)

assert_type(AR_M - AR_LIKE_b, npt.NDArray[np.datetime64])
assert_type(AR_M - AR_LIKE_u, npt.NDArray[np.datetime64])
assert_type(AR_M - AR_LIKE_i, npt.NDArray[np.datetime64])
assert_type(AR_M - AR_LIKE_m, npt.NDArray[np.datetime64])
assert_type(AR_M - AR_LIKE_M, npt.NDArray[np.timedelta64])
assert_type(AR_M - AR_LIKE_O, Any)

assert_type(AR_LIKE_M - AR_M, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_O - AR_M, Any)

assert_type(AR_O - AR_LIKE_b, Any)
assert_type(AR_O - AR_LIKE_u, Any)
assert_type(AR_O - AR_LIKE_i, Any)
assert_type(AR_O - AR_LIKE_f, Any)
assert_type(AR_O - AR_LIKE_c, Any)
assert_type(AR_O - AR_LIKE_m, Any)
assert_type(AR_O - AR_LIKE_M, Any)
assert_type(AR_O - AR_LIKE_O, Any)

assert_type(AR_LIKE_b - AR_O, Any)
assert_type(AR_LIKE_u - AR_O, Any)
assert_type(AR_LIKE_i - AR_O, Any)
assert_type(AR_LIKE_f - AR_O, Any)
assert_type(AR_LIKE_c - AR_O, Any)
assert_type(AR_LIKE_m - AR_O, Any)
assert_type(AR_LIKE_M - AR_O, Any)
assert_type(AR_LIKE_O - AR_O, Any)

# Array floor division

assert_type(AR_b // AR_LIKE_b, npt.NDArray[np.int8])
assert_type(AR_b // AR_LIKE_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_b // AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_b // AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_b // AR_LIKE_O, Any)

assert_type(AR_LIKE_b // AR_b, npt.NDArray[np.int8])
assert_type(AR_LIKE_u // AR_b, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_i // AR_b, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f // AR_b, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_O // AR_b, Any)

assert_type(AR_u // AR_LIKE_b, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_u // AR_LIKE_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_u // AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_u // AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_u // AR_LIKE_O, Any)

assert_type(AR_LIKE_b // AR_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_u // AR_u, npt.NDArray[np.unsignedinteger[Any]])
assert_type(AR_LIKE_i // AR_u, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f // AR_u, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_m // AR_u, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_O // AR_u, Any)

assert_type(AR_i // AR_LIKE_b, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i // AR_LIKE_u, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i // AR_LIKE_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_i // AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_i // AR_LIKE_O, Any)

assert_type(AR_LIKE_b // AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_u // AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_i // AR_i, npt.NDArray[np.signedinteger[Any]])
assert_type(AR_LIKE_f // AR_i, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_m // AR_i, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_O // AR_i, Any)

assert_type(AR_f // AR_LIKE_b, npt.NDArray[np.floating[Any]])
assert_type(AR_f // AR_LIKE_u, npt.NDArray[np.floating[Any]])
assert_type(AR_f // AR_LIKE_i, npt.NDArray[np.floating[Any]])
assert_type(AR_f // AR_LIKE_f, npt.NDArray[np.floating[Any]])
assert_type(AR_f // AR_LIKE_O, Any)

assert_type(AR_LIKE_b // AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_u // AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_i // AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_f // AR_f, npt.NDArray[np.floating[Any]])
assert_type(AR_LIKE_m // AR_f, npt.NDArray[np.timedelta64])
assert_type(AR_LIKE_O // AR_f, Any)

assert_type(AR_m // AR_LIKE_u, npt.NDArray[np.timedelta64])
assert_type(AR_m // AR_LIKE_i, npt.NDArray[np.timedelta64])
assert_type(AR_m // AR_LIKE_f, npt.NDArray[np.timedelta64])
assert_type(AR_m // AR_LIKE_m, npt.NDArray[np.int64])
assert_type(AR_m // AR_LIKE_O, Any)

assert_type(AR_LIKE_m // AR_m, npt.NDArray[np.int64])
assert_type(AR_LIKE_O // AR_m, Any)

assert_type(AR_O // AR_LIKE_b, Any)
assert_type(AR_O // AR_LIKE_u, Any)
assert_type(AR_O // AR_LIKE_i, Any)
assert_type(AR_O // AR_LIKE_f, Any)
assert_type(AR_O // AR_LIKE_m, Any)
assert_type(AR_O // AR_LIKE_M, Any)
assert_type(AR_O // AR_LIKE_O, Any)

assert_type(AR_LIKE_b // AR_O, Any)
assert_type(AR_LIKE_u // AR_O, Any)
assert_type(AR_LIKE_i // AR_O, Any)
assert_type(AR_LIKE_f // AR_O, Any)
assert_type(AR_LIKE_m // AR_O, Any)
assert_type(AR_LIKE_M // AR_O, Any)
assert_type(AR_LIKE_O // AR_O, Any)

# unary ops

assert_type(-f16, np.floating[_128Bit])
assert_type(-c16, np.complex128)
assert_type(-c8, np.complex64)
assert_type(-f8, np.float64)
assert_type(-f4, np.float32)
assert_type(-i8, np.int64)
assert_type(-i4, np.int32)
assert_type(-u8, np.uint64)
assert_type(-u4, np.uint32)
assert_type(-td, np.timedelta64)
assert_type(-AR_f, npt.NDArray[np.float64])

assert_type(+f16, np.floating[_128Bit])
assert_type(+c16, np.complex128)
assert_type(+c8, np.complex64)
assert_type(+f8, np.float64)
assert_type(+f4, np.float32)
assert_type(+i8, np.int64)
assert_type(+i4, np.int32)
assert_type(+u8, np.uint64)
assert_type(+u4, np.uint32)
assert_type(+td, np.timedelta64)
assert_type(+AR_f, npt.NDArray[np.float64])

assert_type(abs(f16), np.floating[_128Bit])
assert_type(abs(c16), np.float64)
assert_type(abs(c8), np.float32)
assert_type(abs(f8), np.float64)
assert_type(abs(f4), np.float32)
assert_type(abs(i8), np.int64)
assert_type(abs(i4), np.int32)
assert_type(abs(u8), np.uint64)
assert_type(abs(u4), np.uint32)
assert_type(abs(td), np.timedelta64)
assert_type(abs(b_), np.bool)

# Time structures

assert_type(dt + td, np.datetime64)
assert_type(dt + i, np.datetime64)
assert_type(dt + i4, np.datetime64)
assert_type(dt + i8, np.datetime64)
assert_type(dt - dt, np.timedelta64)
assert_type(dt - i, np.datetime64)
assert_type(dt - i4, np.datetime64)
assert_type(dt - i8, np.datetime64)

assert_type(td + td, np.timedelta64)
assert_type(td + i, np.timedelta64)
assert_type(td + i4, np.timedelta64)
assert_type(td + i8, np.timedelta64)
assert_type(td - td, np.timedelta64)
assert_type(td - i, np.timedelta64)
assert_type(td - i4, np.timedelta64)
assert_type(td - i8, np.timedelta64)
assert_type(td / f, np.timedelta64)
assert_type(td / f4, np.timedelta64)
assert_type(td / f8, np.timedelta64)
assert_type(td / td, np.float64)
assert_type(td // td, np.int64)

# boolean

assert_type(b_ / b, np.float64)
assert_type(b_ / b_, np.float64)
assert_type(b_ / i, np.float64)
assert_type(b_ / i8, np.float64)
assert_type(b_ / i4, np.float64)
assert_type(b_ / u8, np.float64)
assert_type(b_ / u4, np.float64)
assert_type(b_ / f, np.float64)
assert_type(b_ / f16, np.floating[_128Bit])
assert_type(b_ / f8, np.float64)
assert_type(b_ / f4, np.float32)
assert_type(b_ / c, np.complex128)
assert_type(b_ / c16, np.complex128)
assert_type(b_ / c8, np.complex64)

assert_type(b / b_, np.float64)
assert_type(b_ / b_, np.float64)
assert_type(i / b_, np.float64)
assert_type(i8 / b_, np.float64)
assert_type(i4 / b_, np.float64)
assert_type(u8 / b_, np.float64)
assert_type(u4 / b_, np.float64)
assert_type(f / b_, np.float64)
assert_type(f16 / b_, np.floating[_128Bit])
assert_type(f8 / b_, np.float64)
assert_type(f4 / b_, np.float32)
assert_type(c / b_, np.complex128)
assert_type(c16 / b_, np.complex128)
assert_type(c8 / b_, np.complex64)

# Complex

assert_type(c16 + f16, np.complexfloating[_64Bit | _128Bit, _64Bit | _128Bit])
assert_type(c16 + c16, np.complex128)
assert_type(c16 + f8, np.complex128)
assert_type(c16 + i8, np.complex128)
assert_type(c16 + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c16 + f4, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c16 + i4, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c16 + b_, np.complex128)
assert_type(c16 + b, np.complex128)
assert_type(c16 + c, np.complex128)
assert_type(c16 + f, np.complex128)
assert_type(c16 + AR_f, npt.NDArray[np.complexfloating[Any, Any]])

assert_type(f16 + c16, np.complexfloating[_64Bit | _128Bit, _64Bit | _128Bit])
assert_type(c16 + c16, np.complex128)
assert_type(f8 + c16, np.complex128)
assert_type(i8 + c16, np.complex128)
assert_type(c8 + c16, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(f4 + c16, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(i4 + c16, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(b_ + c16, np.complex128)
assert_type(b + c16, np.complex128)
assert_type(c + c16, np.complex128)
assert_type(f + c16, np.complex128)
assert_type(AR_f + c16, npt.NDArray[np.complexfloating[Any, Any]])

assert_type(c8 + f16, np.complexfloating[_32Bit | _128Bit, _32Bit | _128Bit])
assert_type(c8 + c16, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + f8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + i8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + c8, np.complex64)
assert_type(c8 + f4, np.complex64)
assert_type(c8 + i4, np.complex64)
assert_type(c8 + b_, np.complex64)
assert_type(c8 + b, np.complex64)
assert_type(c8 + c, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + f, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + AR_f, npt.NDArray[np.complexfloating[Any, Any]])

assert_type(f16 + c8, np.complexfloating[_32Bit | _128Bit, _32Bit | _128Bit])
assert_type(c16 + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(f8 + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(i8 + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(c8 + c8, np.complex64)
assert_type(f4 + c8, np.complex64)
assert_type(i4 + c8, np.complex64)
assert_type(b_ + c8, np.complex64)
assert_type(b + c8, np.complex64)
assert_type(c + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(f + c8, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(AR_f + c8, npt.NDArray[np.complexfloating[Any, Any]])

# Float

assert_type(f8 + f16, np.floating[_64Bit | _128Bit])
assert_type(f8 + f8, np.float64)
assert_type(f8 + i8, np.float64)
assert_type(f8 + f4, np.floating[_32Bit | _64Bit])
assert_type(f8 + i4, np.floating[_32Bit | _64Bit])
assert_type(f8 + b_, np.float64)
assert_type(f8 + b, np.float64)
assert_type(f8 + c, np.complex128)
assert_type(f8 + f, np.float64)
assert_type(f8 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(f16 + f8, np.floating[_64Bit | _128Bit])
assert_type(f8 + f8, np.float64)
assert_type(i8 + f8, np.float64)
assert_type(f4 + f8, np.floating[_32Bit | _64Bit])
assert_type(i4 + f8, np.floating[_32Bit | _64Bit])
assert_type(b_ + f8, np.float64)
assert_type(b + f8, np.float64)
assert_type(c + f8, np.complex128)
assert_type(f + f8, np.float64)
assert_type(AR_f + f8, npt.NDArray[np.floating[Any]])

assert_type(f4 + f16, np.floating[_32Bit | _128Bit])
assert_type(f4 + f8, np.floating[_32Bit | _64Bit])
assert_type(f4 + i8, np.floating[_32Bit | _64Bit])
assert_type(f4 + f4, np.float32)
assert_type(f4 + i4, np.float32)
assert_type(f4 + b_, np.float32)
assert_type(f4 + b, np.float32)
assert_type(f4 + c, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(f4 + f, np.floating[_32Bit | _64Bit])
assert_type(f4 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(f16 + f4, np.floating[_32Bit | _128Bit])
assert_type(f8 + f4, np.floating[_32Bit | _64Bit])
assert_type(i8 + f4, np.floating[_32Bit | _64Bit])
assert_type(f4 + f4, np.float32)
assert_type(i4 + f4, np.float32)
assert_type(b_ + f4, np.float32)
assert_type(b + f4, np.float32)
assert_type(c + f4, np.complexfloating[_32Bit | _64Bit, _32Bit | _64Bit])
assert_type(f + f4, np.floating[_32Bit | _64Bit])
assert_type(AR_f + f4, npt.NDArray[np.floating[Any]])

# Int

assert_type(i8 + i8, np.int64)
assert_type(i8 + u8, Any)
assert_type(i8 + i4, np.signedinteger[_32Bit | _64Bit])
assert_type(i8 + u4, Any)
assert_type(i8 + b_, np.int64)
assert_type(i8 + b, np.int64)
assert_type(i8 + c, np.complex128)
assert_type(i8 + f, np.float64)
assert_type(i8 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(u8 + u8, np.uint64)
assert_type(u8 + i4, Any)
assert_type(u8 + u4, np.unsignedinteger[_32Bit | _64Bit])
assert_type(u8 + b_, np.uint64)
assert_type(u8 + b, np.uint64)
assert_type(u8 + c, np.complex128)
assert_type(u8 + f, np.float64)
assert_type(u8 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(i8 + i8, np.int64)
assert_type(u8 + i8, Any)
assert_type(i4 + i8, np.signedinteger[_32Bit | _64Bit])
assert_type(u4 + i8, Any)
assert_type(b_ + i8, np.int64)
assert_type(b + i8, np.int64)
assert_type(c + i8, np.complex128)
assert_type(f + i8, np.float64)
assert_type(AR_f + i8, npt.NDArray[np.floating[Any]])

assert_type(u8 + u8, np.uint64)
assert_type(i4 + u8, Any)
assert_type(u4 + u8, np.unsignedinteger[_32Bit | _64Bit])
assert_type(b_ + u8, np.uint64)
assert_type(b + u8, np.uint64)
assert_type(c + u8, np.complex128)
assert_type(f + u8, np.float64)
assert_type(AR_f + u8, npt.NDArray[np.floating[Any]])

assert_type(i4 + i8, np.signedinteger[_32Bit | _64Bit])
assert_type(i4 + i4, np.int32)
assert_type(i4 + b_, np.int32)
assert_type(i4 + b, np.int32)
assert_type(i4 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(u4 + i8, Any)
assert_type(u4 + i4, Any)
assert_type(u4 + u8, np.unsignedinteger[_32Bit | _64Bit])
assert_type(u4 + u4, np.uint32)
assert_type(u4 + b_, np.uint32)
assert_type(u4 + b, np.uint32)
assert_type(u4 + AR_f, npt.NDArray[np.floating[Any]])

assert_type(i8 + i4, np.signedinteger[_32Bit | _64Bit])
assert_type(i4 + i4, np.int32)
assert_type(b_ + i4, np.int32)
assert_type(b + i4, np.int32)
assert_type(AR_f + i4, npt.NDArray[np.floating[Any]])

assert_type(i8 + u4, Any)
assert_type(i4 + u4, Any)
assert_type(u8 + u4, np.unsignedinteger[_32Bit | _64Bit])
assert_type(u4 + u4, np.uint32)
assert_type(b_ + u4, np.uint32)
assert_type(b + u4, np.uint32)
assert_type(AR_f + u4, npt.NDArray[np.floating[Any]])



from __future__ import annotations

import importlib.util
import os
import re
import shutil
from collections import defaultdict
from collections.abc import Iterator
from typing import TYPE_CHECKING

import pytest
from numpy.typing.mypy_plugin import _EXTENDED_PRECISION_LIST


# Only trigger a full `mypy` run if this environment variable is set
# Note that these tests tend to take over a minute even on a macOS M1 CPU,
# and more than that in CI.
RUN_MYPY = "NPY_RUN_MYPY_IN_TESTSUITE" in os.environ
if RUN_MYPY and RUN_MYPY not in ('0', '', 'false'):
    RUN_MYPY = True

# Skips all functions in this file
pytestmark = pytest.mark.skipif(
    not RUN_MYPY,
    reason="`NPY_RUN_MYPY_IN_TESTSUITE` not set"
)


try:
    from mypy import api
except ImportError:
    NO_MYPY = True
else:
    NO_MYPY = False

if TYPE_CHECKING:
    # We need this as annotation, but it's located in a private namespace.
    # As a compromise, do *not* import it during runtime
    from _pytest.mark.structures import ParameterSet

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
PASS_DIR = os.path.join(DATA_DIR, "pass")
FAIL_DIR = os.path.join(DATA_DIR, "fail")
REVEAL_DIR = os.path.join(DATA_DIR, "reveal")
MISC_DIR = os.path.join(DATA_DIR, "misc")
MYPY_INI = os.path.join(DATA_DIR, "mypy.ini")
CACHE_DIR = os.path.join(DATA_DIR, ".mypy_cache")

#: A dictionary with file names as keys and lists of the mypy stdout as values.
#: To-be populated by `run_mypy`.
OUTPUT_MYPY: defaultdict[str, list[str]] = defaultdict(list)


def _key_func(key: str) -> str:
    """Split at the first occurrence of the ``:`` character.

    Windows drive-letters (*e.g.* ``C:``) are ignored herein.
    """
    drive, tail = os.path.splitdrive(key)
    return os.path.join(drive, tail.split(":", 1)[0])


def _strip_filename(msg: str) -> tuple[int, str]:
    """Strip the filename and line number from a mypy message."""
    _, tail = os.path.splitdrive(msg)
    _, lineno, msg = tail.split(":", 2)
    return int(lineno), msg.strip()


def strip_func(match: re.Match[str]) -> str:
    """`re.sub` helper function for stripping module names."""
    return match.groups()[1]


@pytest.fixture(scope="module", autouse=True)
def run_mypy() -> None:
    """Clears the cache and run mypy before running any of the typing tests.

    The mypy results are cached in `OUTPUT_MYPY` for further use.

    The cache refresh can be skipped using

    NUMPY_TYPING_TEST_CLEAR_CACHE=0 pytest numpy/typing/tests
    """
    if (
        os.path.isdir(CACHE_DIR)
        and bool(os.environ.get("NUMPY_TYPING_TEST_CLEAR_CACHE", True))
    ):
        shutil.rmtree(CACHE_DIR)

    split_pattern = re.compile(r"(\s+)?\^(\~+)?")
    for directory in (PASS_DIR, REVEAL_DIR, FAIL_DIR, MISC_DIR):
        # Run mypy
        stdout, stderr, exit_code = api.run([
            "--config-file",
            MYPY_INI,
            "--cache-dir",
            CACHE_DIR,
            directory,
        ])
        if stderr:
            pytest.fail(f"Unexpected mypy standard error\n\n{stderr}")
        elif exit_code not in {0, 1}:
            pytest.fail(f"Unexpected mypy exit code: {exit_code}\n\n{stdout}")

        str_concat = ""
        filename: str | None = None
        for i in stdout.split("\n"):
            if "note:" in i:
                continue
            if filename is None:
                filename = _key_func(i)

            str_concat += f"{i}\n"
            if split_pattern.match(i) is not None:
                OUTPUT_MYPY[filename].append(str_concat)
                str_concat = ""
                filename = None


def get_test_cases(directory: str) -> Iterator[ParameterSet]:
    for root, _, files in os.walk(directory):
        for fname in files:
            short_fname, ext = os.path.splitext(fname)
            if ext in (".pyi", ".py"):
                fullpath = os.path.join(root, fname)
                yield pytest.param(fullpath, id=short_fname)


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(PASS_DIR))
def test_success(path) -> None:
    # Alias `OUTPUT_MYPY` so that it appears in the local namespace
    output_mypy = OUTPUT_MYPY
    if path in output_mypy:
        msg = "Unexpected mypy output\n\n"
        msg += "\n".join(_strip_filename(v)[1] for v in output_mypy[path])
        raise AssertionError(msg)


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(FAIL_DIR))
def test_fail(path: str) -> None:
    __tracebackhide__ = True

    with open(path) as fin:
        lines = fin.readlines()

    errors = defaultdict(lambda: "")

    output_mypy = OUTPUT_MYPY
    assert path in output_mypy

    for error_line in output_mypy[path]:
        lineno, error_line = _strip_filename(error_line)
        errors[lineno] += f'{error_line}\n'

    for i, line in enumerate(lines):
        lineno = i + 1
        if (
            line.startswith('#')
            or (" E:" not in line and lineno not in errors)
        ):
            continue

        target_line = lines[lineno - 1]
        if "# E:" in target_line:
            expression, _, marker = target_line.partition("  # E: ")
            expected_error = errors[lineno].strip()
            marker = marker.strip()
            _test_fail(path, expression, marker, expected_error, lineno)
        else:
            pytest.fail(
                f"Unexpected mypy output at line {lineno}\n\n{errors[lineno]}"
            )


_FAIL_MSG1 = """Extra error at line {}

Expression: {}
Extra error: {!r}
"""

_FAIL_MSG2 = """Error mismatch at line {}

Expression: {}
Expected error: {}
Observed error: {!r}
"""


def _test_fail(
    path: str,
    expression: str,
    error: str,
    expected_error: None | str,
    lineno: int,
) -> None:
    if expected_error is None:
        raise AssertionError(_FAIL_MSG1.format(lineno, expression, error))
    elif error not in expected_error:
        raise AssertionError(_FAIL_MSG2.format(
            lineno, expression, expected_error, error
        ))


_REVEAL_MSG = """Reveal mismatch at line {}

{}
"""


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(REVEAL_DIR))
def test_reveal(path: str) -> None:
    """Validate that mypy correctly infers the return-types of
    the expressions in `path`.
    """
    __tracebackhide__ = True

    output_mypy = OUTPUT_MYPY
    if path not in output_mypy:
        return

    for error_line in output_mypy[path]:
        lineno, error_line = _strip_filename(error_line)
        raise AssertionError(_REVEAL_MSG.format(lineno, error_line))


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
@pytest.mark.parametrize("path", get_test_cases(PASS_DIR))
def test_code_runs(path: str) -> None:
    """Validate that the code in `path` properly during runtime."""
    path_without_extension, _ = os.path.splitext(path)
    dirname, filename = path.split(os.sep)[-2:]

    spec = importlib.util.spec_from_file_location(
        f"{dirname}.{filename}", path
    )
    assert spec is not None
    assert spec.loader is not None

    test_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(test_module)


LINENO_MAPPING = {
    11: "uint128",
    12: "uint256",
    14: "int128",
    15: "int256",
    17: "float80",
    18: "float96",
    19: "float128",
    20: "float256",
    22: "complex160",
    23: "complex192",
    24: "complex256",
    25: "complex512",
}


@pytest.mark.slow
@pytest.mark.skipif(NO_MYPY, reason="Mypy is not installed")
def test_extended_precision() -> None:
    path = os.path.join(MISC_DIR, "extended_precision.pyi")
    output_mypy = OUTPUT_MYPY
    assert path in output_mypy

    with open(path) as f:
        expression_list = f.readlines()

    for _msg in output_mypy[path]:
        lineno, msg = _strip_filename(_msg)
        expression = expression_list[lineno - 1].rstrip("\n")

        if LINENO_MAPPING[lineno] in _EXTENDED_PRECISION_LIST:
            raise AssertionError(_REVEAL_MSG.format(lineno, msg))
        elif "error" not in msg:
            _test_fail(
                path, expression, msg, 'Expression is of type "Any"', lineno
            )



from numpy import (
    array as _np_array, 
    unique, linspace, invert, flip, transpose, concatenate, sign, zeros, 
    min as _np_min, max as _np_max, equal, where, logical_or, logical_and, all, newaxis)

def apply_invalid_area_on_grid(
        
):
    return


def disk_cell_intersection_area(
    disk_center_pt:_np_array,      
    cell:float=0.0075,
    grid_spacing:float=0.0075,
    r:float=0.0075,
    silent = False,
):
    """
    Returns intersection area. area if intesection (0,grid_spacing**2)
    Case for no intersection will be handled before (fully included or fully excluded).
    Case 1: two intersection points (more than half of square are within radius) - 3 vertices are within circle 
    Case 2: two intersection points (more than half of square are within radius) - 1 vertex is within circle
    Case 3: two intersection points (less than half of square within radius) - 0 vertices within circle (same row or col)
    Case 4: two intersection points (unclear wheter more or less than half) - 2 vertices within circle (same row or col)
    Case 5: four intersection points (more than half of circle is included) - 2 vertices within circle (same row or col)
    
    TODO: if grid_spacing/2 is greater than radius there will be weird instances 

    This can also be done already as a function of the point offset
    and is also symmetrical towards the triangle
    the intersection area only needs to be computed for those cases where excluded cells are intersected 

    """
    # case no intersection will be handled
    area = grid_spacing**2
    return area




from numpy import (
    array as _np_array,
    append as _np_append,
    invert as _np_invert,
    sign as _np_sign,
    all as _np_all,
    any as _np_any,
    zeros as _np_zeros,
    ones as _np_ones,
    dstack as _np_dstack,
    logical_or,
    where as _np_where, 
    arange, transpose as _np_transpose, unique, linspace, flip,  concatenate,
    min,
    max as _np_max, 
    equal, logical_and, 
    newaxis as _np_newaxis, 
    sort
)
from numpy.linalg import norm as _np_linalg_norm
from math import ceil as _math_ceil
from ..utils.general import ( flatten_list, visualize, depth, list_dict_keys, )


# TODO vectorize them such that they can take any input from shape(1,2) to shape(n,2)
# TODO add descriptions
def min_possible_dist_cells_to_cell(
        cell_orig:_np_array,
        cell_dest:_np_array=_np_array([0,0]),
        ) -> float:
    """
    TODO
    Returns:

    """
    if len(cell_orig.shape)==1:
        cell_orig = cell_orig[_np_newaxis]

    cell_diff = cell_orig - cell_dest
    
    return _np_linalg_norm(cell_diff - _np_sign(cell_diff), axis=1)
    # return grid_spacing*((row_diff-_np_sign(row_diff))**2 + (col_diff-_np_sign(col_diff))**2)**.5

#
def max_possible_dist_cells_to_cell(
        cell_dest:_np_array,
        cell_orig:_np_array=_np_array([0,0]),
    ) -> float:
    """
    TODO
    Returns:

    """
    if len(cell_orig.shape)==1:
        cell_orig = cell_orig[_np_newaxis]
        
    cell_diff = cell_orig - cell_dest
    return _np_linalg_norm(cell_diff + _np_where(cell_diff>=0,1,-1), axis=1)
    # return grid_spacing*((row_diff+(1 if row_diff>=0 else -1))**2 + (col_diff+(1 if col_diff>=0 else -1))**2)**.5
#
def min_possible_dist_trgl1_to_cell(
        cell_dest:_np_array,
        ) -> float:
    """
    TODO
    TODO link triangle 1 definition
    Returns:

    """
    (row, col) = cell_dest

    # return grid_spacing*((row-(.5 if row>0 else 0))**2 + (col-(.5 if col>0 else 0))**2)**.5
    return (
        (row-(1 if row>0 else 0 if row==0 else -0.5))**2 + 
        (col-(1 if col>0 else 0 if col==0 else -0.5))**2
        )**.5
#
def max_possible_dist_trgl1_to_cell(
        cell_dest:_np_array,
        ) -> float:
    """
    TODO
    unscaled
    Returns:

    """
    (row, col) = cell_dest
    return (
        (row+(0.5 if row>=0 else -1))**2 + 
        (col+(0.5 if col>=0 else -1))**2
        )**.5
#

def get_cell_closest_point_to_points(
        pts_xy:_np_array,
        cell:_np_array,
        ):
    """
    TODO THIS MATTER FOR CELL REGIONS, 
    THIS COULD ALSO RETURN LINES FOR POINTS  
    Returns:
    """
    # TODO also check if handle correctly case for same row / column
    pts_x_unscaled = pts_xy[:, 0]
    pts_y_unscaled = pts_xy[:,1]
    n_pts = pts_xy.shape[0]

    row,col = cell
    ys = (
        # TODO this could also account for cases where point is not within cell (0,0)
        pts_y_unscaled
    ) if row == 0 else (
        # pts_y_unscaled - (row - .5 * _np_sign(row))
       _np_zeros(n_pts) + row - .5 * _np_sign(row)
    )
    
    xs = (
        # TODO this could also account for cases where point is not within cell (0,0)
        pts_x_unscaled
    ) if col == 0 else (
        # pts_x_unscaled - (col - .5 * _np_sign(col))
        _np_zeros(n_pts) + col - .5 * _np_sign(col)
    )

    return _np_transpose([xs, ys], (1,0))
#

def get_cell_closest_point_to_point(
        pt_xy:_np_array,
        cell:_np_array,
        ):
    """
    TODO THIS MATTER FOR CELL REGIONS, 
    THIS COULD ALSO RETURN LINES FOR POINTS  
    Returns:
    """
    # TODO also check if handle correctly case for same row / column
    x,y = pt_xy
    row,col = cell
    return _np_array([
        x if col == 0 else col - .5 * _np_sign(col),
        y if row == 0 else row - .5 * _np_sign(row)
    ])
#

def min_dist_points_to_cell(
        pts_xy:_np_array,
        cell:_np_array,
        ) -> _np_array:
    """
    TODO
    TODO link triangle 1 definition
    Returns:
    ONLYACCEPTS POINTS WITHIN CELL (0,0)!
    POINTS MUST BE RELATIVE TO GRID [-0.5,0.5]
    """

    # TODO also check if handle correctly case for same row / column
    pts_x_unscaled = pts_xy[:, 0]
    pts_y_unscaled = pts_xy[:,1]
    n = pts_xy.shape[0]

    row,col = cell
    
    ydist = (
        # TODO this could also account for cases where point is not within cell (0,0)
        _np_zeros(n)
    ) if row == 0 else (
        pts_y_unscaled - (row - .5 * _np_sign(row))
    )
    
    xdist = (
        # TODO this could also account for cases where point is not within cell (0,0)
        _np_zeros(n)
    ) if col == 0 else (
        pts_x_unscaled - (col - .5 * _np_sign(col))
    )
    
    return _np_linalg_norm([xdist, ydist], axis=0)
#


def get_cell_farthest_vertex_to_point(
        pts_xy:_np_array,
        cell:_np_array,
        ):
    """
    TODO THIS MATTER FOR CELL REGIONS, 
    THIS COULD ALSO RETURN LINES FOR POINTS  
    Returns:

    """
    return (cell[::-1] + (cell[::-1] - pts_xy >= 0)-.5)
#

def max_dist_points_to_cell(
        pts_xy:_np_array,
        cell:_np_array,
        ) -> _np_array:
    """
    TODO
    Returns:
        1D array of min distance from points to cell
    """
    if len(pts_xy)==0:
        print("NOOOOPOINT1",pts_xy)
        return _np_array([])
    if len(cell)==0:
        print("NOOOOPOINTcell_cell",cell)
    return _np_linalg_norm(
        _np_array([
            get_cell_farthest_vertex_to_point(
                pts_xy,
                cell
            ) for pts_xy in pts_xy
            ]) - pts_xy,
        axis=1
    )
#

def is_within_radius(distances,r:float,include_boundary):
    if include_boundary:
        return distances <= r
    return distances < r

def get_cells_relevant_for_disk_by_type(
        grid_spacing:float=250,
        r:float=750,
        include_boundary:bool=False,
        return_type:str=[None, 'contained_by_all', 'contained_by_trgl', 'overlapped_by_all', 'overlapped_by_trgl'][0]
    ) -> tuple:
    """
    TODO turn it into smaller function to return a specific type only 
    For an undefined point in cell(0,0): Determines which cells are always fully within radius and which cell might be overlaped 
    If count == True returns only count of cells. If not True returns list of tuples of cell row col locations
    Note due to symmetry only 1/8 of radius needs to be checked 
    Returns
    - cells_contained_in_all_disks, 
    - cells_contained_in_all_trgl_disks, 
    - cells_maybe_overlapping_a_disk, 
    - cells_maybe_overlapping_a_trgl_disk
    """

    ratio = r/grid_spacing
    cell_steps_max = _math_ceil(ratio+1)
    
    # create all cells within square
    unassigned_cells_in_max_steps_square = _np_array(flatten_list([[(row,col) for col in range(-cell_steps_max, cell_steps_max+1)] for row in range(-cell_steps_max, cell_steps_max+1)]))
    # compare farthest vertex of cells: bool if cell is in contained in disk of radius centered around any pt within grid cell 
    cell_is_contained_in_all_disks = grid_spacing * max_possible_dist_cells_to_cell(unassigned_cells_in_max_steps_square) <= radius
    cells_contained_in_all_disks = unassigned_cells_in_max_steps_square[cell_is_contained_in_all_disks,:]
    
    if 'contained_by_all' == return_type:
        return cells_contained_in_all_disks
    
    # those cells should not be part of other sets
    unassigned_cells_in_max_steps_square = unassigned_cells_in_max_steps_square[_np_invert(cell_is_contained_in_all_disks),:]
    # compare closest vertex of cells: bool if cell is potentially overlaped by a disk of radius around pt within grid cell 
    cell_may_overlap_a_disk = (
        grid_spacing * min_possible_dist_cells_to_cell(unassigned_cells_in_max_steps_square) <= r
        ) if include_boundary else (
        grid_spacing * min_possible_dist_cells_to_cell(unassigned_cells_in_max_steps_square) < r
        )
    # [min_possible_dist_cell_to_cell(cell, grid_spacing) <= r for cell in unassigned_cells_in_max_steps_square]
    cells_maybe_overlapping_a_disk = unassigned_cells_in_max_steps_square[cell_may_overlap_a_disk,:]
    if 'overlapped_by_all' == return_type:
        return cells_maybe_overlapping_a_disk
    
    # compare farthest vertex of cells: bool if cell is in contained in disk of radius centered around any pt within triangle 1 
    cell_is_contained_in_all_trgl_disks = [grid_spacing * max_possible_dist_trgl1_to_cell(cell) <= r for cell in unassigned_cells_in_max_steps_square]
    cells_contained_in_all_trgl_disks = unassigned_cells_in_max_steps_square[cell_is_contained_in_all_trgl_disks]    
    
    if 'contained_by_trgl' == return_type:
        return cells_contained_in_all_trgl_disks
    
    # those cells should not be in cells maybe overlapping disk around pts in triangle as they are surely contain
    unassigned_cells_in_max_steps_square = unassigned_cells_in_max_steps_square[_np_invert(cell_is_contained_in_all_trgl_disks),:]
    # compare closest vertex of cells: bool if cell is potentially overlaped by a disk of radius around pt within triangle 1
    cell_may_overlap_trgl_disk = (
        [grid_spacing * min_possible_dist_trgl1_to_cell(cell) <= r for cell in unassigned_cells_in_max_steps_square]
        ) if include_boundary else (
        [grid_spacing * min_possible_dist_trgl1_to_cell(cell) < r for cell in unassigned_cells_in_max_steps_square]
        )
    cells_maybe_overlapping_a_trgl_disk = unassigned_cells_in_max_steps_square[cell_may_overlap_trgl_disk,:]

    if 'overlapped_by_trgl' == return_type:
        return cells_maybe_overlapping_a_trgl_disk
    
    return (
        cells_contained_in_all_disks, 
        cells_contained_in_all_trgl_disks, 
        cells_maybe_overlapping_a_disk, 
        cells_maybe_overlapping_a_trgl_disk
    )
#

# relevant cell types:
# cells always contained

def filter_cells_alw_contained(
        grid_spacing:float=250,
        r:float=750,
        include_boundary:bool=False
    ) -> tuple:
    return get_cells_relevant_for_disk_by_type(
       grid_spacing=grid_spacing, r=r, include_boundary=include_boundary, return_type='contained_by_all'
    )
#

def filter_cells_alw_contained(
        grid_spacing:float=250,
        r:float=750,
        include_boundary:bool=False
    ) -> tuple:
    return get_cells_relevant_for_disk_by_type(
       grid_spacing=grid_spacing, r=r, include_boundary=include_boundary, return_type='contained_by_all'
    )
#

def filter_cells_alw_overlapped(
        grid_spacing:float=250,
        r:float=750,
        include_boundary:bool=False
    ) -> tuple:
    return get_cells_relevant_for_disk_by_type(
       grid_spacing=grid_spacing, r=r, include_boundary=include_boundary, return_type='overlapped_by_trgl'
    )
#

def filter_cells_alw_overlapped(
        grid_spacing:float=250,
        r:float=750,
        include_boundary:bool=False
    ) -> tuple:
    return get_cells_relevant_for_disk_by_type(
       grid_spacing=grid_spacing, r=r, include_boundary=include_boundary, return_type='overlapped_by_trgl'
    )
#

def check_if_always_overlaps_convex_set(
        cells:_np_array,
        convex_set_vertices:_np_array = _np_array([[0,0],[.5,0],[.5,.5]]),
        vertex_is_inside_convex_set=True,
        r:float=0.00750,
        grid_spacing:float=0.00250, 
        include_boundary:bool=False,
):
    if type(vertex_is_inside_convex_set)==bool:
        if vertex_is_inside_convex_set:
            vertex_is_inside_convex_set = _np_ones(len(convex_set_vertices), dtype=bool)
        else:
            vertex_is_inside_convex_set = _np_zeros(len(convex_set_vertices), dtype=bool)
    
    # smallest possible distance of cell to convex set
    dist_lower_bounds = grid_spacing * _np_array([
        min_dist_points_to_cell(convex_set_vertices,cell) 
        for cell in cells]
    )

    set_always_overlaps_cells = _np_all([
            (dist_lower_bounds[:,i] <= r) if include_boundary and point_in_set or 0. in cell else (dist_lower_bounds[:,i] < r) 
            for i,point_in_set, cell in zip(range(dist_lower_bounds.shape[1]), vertex_is_inside_convex_set, cells)
        ], axis=0)
    
    return set_always_overlaps_cells
#
  
def check_if_always_overlaps_full_convex_set(
        cells:_np_array,
        convex_set_vertices:_np_array = _np_array([[0,0],[.5,0],[.5,.5]]),
        vertex_is_inside_convex_set=True,
        r:float=0.00750,
        grid_spacing:float=0.00250,
        include_boundary:bool=False,
):
    if type(vertex_is_inside_convex_set)==bool:
        if vertex_is_inside_convex_set:
            vertex_is_inside_convex_set = _np_ones(len(convex_set_vertices), dtype=bool)
        else:
            vertex_is_inside_convex_set = _np_zeros(len(convex_set_vertices), dtype=bool)
    
    always_overlaps_full_convex_set = [max([
        (pt_x-vtx_x)**2+(pt_y-vtx_y)**2 for (pt_x, pt_y), vtx_x, vtx_y in [
            (get_cell_closest_point_to_point((vtx_x, vtx_y), cell), vtx_x, vtx_y) for (vtx_x, vtx_y) in convex_set_vertices
        ]
        ]) <= (r/grid_spacing)**2 
        for cell in cells]

    return always_overlaps_full_convex_set
#
  
def check_if_never_contains_convex_set(
        cells:_np_array,
        convex_set_vertices:_np_array = _np_array([[0,0],[.5,0],[.5,.5]]),
        vertex_is_inside_convex_set=True,
        r:float=0.00750,
        grid_spacing:float=0.00250,
        include_boundary:bool=False,
):
    if type(vertex_is_inside_convex_set)==bool:
        if vertex_is_inside_convex_set:
            vertex_is_inside_convex_set = _np_ones(len(convex_set_vertices), dtype=bool)
        else:
            vertex_is_inside_convex_set = _np_zeros(len(convex_set_vertices), dtype=bool)
    
    # largest possible distance of cell to convex set
    dist_upper_bounds = grid_spacing * _np_array([
        max_dist_points_to_cell(convex_set_vertices,cell)
        for cell in cells]
    )

    set_maybe_contains_cells = _np_any(dist_upper_bounds <= r, axis=1)
    set_maybe_contains_cells = _np_any([
            (dist_upper_bounds[:,i] <= r) if include_boundary and point_in_set else (dist_upper_bounds[:,i] < r) 
            for i,point_in_set in zip(range(dist_upper_bounds.shape[1]), vertex_is_inside_convex_set)
        ], axis=0)
    set_never_contains_cells = _np_invert(set_maybe_contains_cells)

    return set_never_contains_cells
#









































def get_always_contained_potentially_overlapped_cells(
        r:float=750,
        grid_spacing:float=750,
        count_only:bool=True
    ) -> tuple:
    """
    TODO this method and get_cells_relevant_for_disk_by_type contain repeated logic. Only one shall be kept.
    For any point within cell in grid (at row=0,col=0) of spacing=grid_spacing=cell width=cell height:
    Determines which cells (relative to the cell where the point falls into) are:
    - contained within radius for all possible points in the cell
    - which cells might be overlaped, i.e. that are not fully contained or partly 
      overlappend for at least one BUT NOT all possible points in the cell
    If count == True returns only count of cells. If not True returns list of tuples of cell row col locations
    Note due to symmetry only 1/8 of radius needs to be checked 
    TODO wording needed to say home grid cell

    Args:
      r (float):
        the radius of disk drawn around any point within the cell, in which other points are searched,
        this is used to check whether other grid cells are fully contained in that disk or overlaped
        by any possble point in the grid cell
      grid_spacing (float):
        the spacing of the grid defining the width and heigth of each grid cell
    
    Returns:
      cell_info (tuple):
        tuple containg:

          - overlap_count (int):
            Number of cells that are contained within disk around all point in cell
          - overlap_count (int):
            Number of cells that are overlap with disk around a point in cell but 
            not contained in disk around all points of cell
        
        and if not count_only additionally containing:
          - overlap_ids (list):
            relative position of cells that are contained within disk around all point in cell
          - overlap_ids (list):
            relative position of cells that are overlap with disk around a point in cell but 
            not contained in disk around all points of cell
          - cell_steps_max (int):
            maximum row (=column) distance of a cell to be potentially overlaped by a disk of
            radius around any point in cell 
     
     """

    ratio = r/grid_spacing
    cell_steps_max = _math_ceil(ratio+1)
    overlap_count, overlap_count = 0,0

    if not count_only:
        overlap_ids, overlap_ids = [], []
    
    for i in range(cell_steps_max):
        for j in range(i+1):
            if ratio > (max(i-1,0)**2+max(j-1,0)**2)**.5:
                # account for symmetry. 
                # Center cell exist once. 
                # 90Deg Diagonals exist 4times. 
                # 45Deg Diagonals exist 4times.
                # Gaps exist 8times.  
                cellCountMultiplier = 1 if (i==0 and j==0) else 4 if (j==0 or i==j) else 8
                i_ids = list(set([-i,i]))
                j_ids = list(set([-j,j]))
                cell_ids = flatten_list([
                    flatten_list([
                        [(i_id,j_id)]+([(j_id,i_id)] if i!=j else []) for j_id in j_ids
                        ]) for i_id in i_ids
                    ])
                n_cells = len(cell_ids)
           
                if ratio >= ((1+i)**2+(1+j)**2)**.5:
                    if not count_only:
                        overlap_ids += cell_ids
                    overlap_count += cellCountMultiplier
                else:
                    if not count_only:
                        overlap_ids += cell_ids
                    overlap_count += cellCountMultiplier
    
    # multiply to get n for full circle not only quadrant
    # overlap_count = max((overlap_count-1)*8+1,0)
    # overlap_count = overlap_count*8

    if count_only:
        return overlap_count, overlap_count

    return overlap_count, overlap_count, sorted(overlap_ids), sorted(overlap_ids), cell_steps_max
#


from typing import Any

import numpy as np
import numpy.typing as npt

AR_f8: npt.NDArray[np.float64]
AR_c16: npt.NDArray[np.complex128]
AR_m: npt.NDArray[np.timedelta64]
AR_M: npt.NDArray[np.datetime64]
AR_O: npt.NDArray[np.object_]

def func(a: int) -> None: ...

np.average(AR_m)  # E: incompatible type
np.select(1, [AR_f8])  # E: incompatible type
np.angle(AR_m)  # E: incompatible type
np.unwrap(AR_m)  # E: incompatible type
np.unwrap(AR_c16)  # E: incompatible type
np.trim_zeros(1)  # E: incompatible type
np.place(1, [True], 1.5)  # E: incompatible type
np.vectorize(1)  # E: incompatible type
np.place(AR_f8, slice(None), 5)  # E: incompatible type

np.interp(AR_f8, AR_c16, AR_f8)  # E: incompatible type
np.interp(AR_c16, AR_f8, AR_f8)  # E: incompatible type
np.interp(AR_f8, AR_f8, AR_f8, period=AR_c16)  # E: No overload variant
np.interp(AR_f8, AR_f8, AR_O)  # E: incompatible type

np.cov(AR_m)  # E: incompatible type
np.cov(AR_O)  # E: incompatible type
np.corrcoef(AR_m)  # E: incompatible type
np.corrcoef(AR_O)  # E: incompatible type
np.corrcoef(AR_f8, bias=True)  # E: No overload variant
np.corrcoef(AR_f8, ddof=2)  # E: No overload variant
np.blackman(1j)  # E: incompatible type
np.bartlett(1j)  # E: incompatible type
np.hanning(1j)  # E: incompatible type
np.hamming(1j)  # E: incompatible type
np.hamming(AR_c16)  # E: incompatible type
np.kaiser(1j, 1)  # E: incompatible type
np.sinc(AR_O)  # E: incompatible type
np.median(AR_M)  # E: incompatible type

np.percentile(AR_f8, 50j)  # E: No overload variant
np.percentile(AR_f8, 50, interpolation="bob")  # E: No overload variant
np.quantile(AR_f8, 0.5j)  # E: No overload variant
np.quantile(AR_f8, 0.5, interpolation="bob")  # E: No overload variant
np.meshgrid(AR_f8, AR_f8, indexing="bob")  # E: incompatible type
np.delete(AR_f8, AR_f8)  # E: incompatible type
np.insert(AR_f8, AR_f8, 1.5)  # E: incompatible type
np.digitize(AR_f8, 1j)  # E: No overload variant



from typing import Any, Iterable, Tuple

from numpy import generic
from numpy.typing import NDArray

__all__: list[str]

# NOTE: In practice `byte_bounds` can (potentially) take any object
# implementing the `__array_interface__` protocol. The caveat is
# that certain keys, marked as optional in the spec, must be present for
#  `byte_bounds`. This concerns `"strides"` and `"data"`.
def byte_bounds(a: generic | NDArray[Any]) -> tuple[int, int]: ...

def normalize_axis_tuple(
    axis: int | Iterable[int], 
    ndim: int = ..., 
    argname: None | str = ..., 
    allow_duplicate: None | bool = ...,
) -> Tuple[int, int]: ...

def normalize_axis_index(
    axis: int = ...,
    ndim: int = ...,
    msg_prefix: None | str = ...,
) -> int: ...



from numpy import (unique as _np_unique, ones as _np_ones,)
from matplotlib.pyplot import (get_cmap as _plt_get_cmap, subplots as _plt_subplots)
from matplotlib.patches import Rectangle as _plt_Rectangle

def illustrate_nested_grid(
            grid:dict,
            pts_lat_lon,
            cell_ids,
            y_min:float=None,
            y_max:float=None,
            x_min:float=None,
            x_max:float=None,
):

    """
    Illustrate method
    """
    # unpack vals
    grid_spacing = grid.spacing
    nested_grid = grid['nested']

    # filter pts to be inside bounds:
    filter_mask = _np_ones(len(pts_lat_lon),dtype=bool) 
    if y_min != None:
        filter_mask = filter_mask * (pts_lat_lon[:,0]>=y_min)
    if y_max != None:
        filter_mask = filter_mask * (pts_lat_lon[:,0]<=y_max)
    if x_min != None:
        filter_mask = filter_mask * (pts_lat_lon[:,1]>=x_min)
    if x_max != None:
        filter_mask = filter_mask * (pts_lat_lon[:,1]<=x_max)
    
    pts_lat_lon = pts_lat_lon[filter_mask] 
    cell_ids = cell_ids[filter_mask]
    print('N Pts:',len(cell_ids), 'in', len(_np_unique(cell_ids)),'cells.')

    def unpackNestedGrid(cell_dict:dict, nest_lvl:int=0):
        """
        TODO Check to move it outside as independent functions
        """
        res = []
        if 'quadrants' in cell_dict:
            for quadrant in cell_dict['quadrants']:
                res += unpackNestedGrid(cell_dict=quadrant,nest_lvl=nest_lvl+1) 
        res += [(nest_lvl, tuple([*tuple([*cell_dict['bounds']])]))]
        return res
    
    flat_nested_grid = []
    for cell_id in _np_unique(cell_ids):
        flat_nested_grid += unpackNestedGrid(nested_grid[cell_id],nest_lvl=0)

    flat_nested_grid = sorted(flat_nested_grid)
    max_nest_level = flat_nested_grid[-1][0]
    print('max_nest_level',max_nest_level)
    cmp = _plt_get_cmap("YlOrBr",max_nest_level)
    # print(flat_nested_grid)
    nested_patches = [_plt_Rectangle(
                # x         y           width           height
                (x_min, y_min), x_max-x_min, y_max-y_min, 
                linewidth=.7, facecolor=(cmp(nest_lvl) if max_nest_level>0 else 'grey'), edgecolor='#444', alpha=1
                ) for nest_lvl, ((y_min, x_min), (y_max, x_max) ) in flat_nested_grid]
    
        
    fig, ax = _plt_subplots(figsize=(15,10))

    for patchToAdd in nested_patches:
        ax.add_patch(patchToAdd)

    ax.scatter(x=pts_lat_lon[:,1], y=pts_lat_lon[:,0], s=0.2,color='black')
    x_steps = grid.x_steps
    y_steps = grid.y_steps
    ax.set_xticks(x_steps)
    ax.set_yticks(y_steps)
    ax.set_xlim(left=x_min, right=x_max)
    ax.set_ylim(bottom=y_min, top=y_max)
    ax.set_aspect('equal', adjustable='box')
    ax.grid()
#


from typing import Any

from numpy import int_
from numpy.typing import NDArray
from numpy.polynomial._polybase import ABCPolyBase
from numpy.polynomial.polyutils import trimcoef

__all__: list[str]

legtrim = trimcoef

def poly2leg(pol): ...
def leg2poly(c): ...

legdomain: NDArray[int_]
legzero: NDArray[int_]
legone: NDArray[int_]
legx: NDArray[int_]

def legline(off, scl): ...
def legfromroots(roots): ...
def legadd(c1, c2): ...
def legsub(c1, c2): ...
def legmulx(c): ...
def legmul(c1, c2): ...
def legdiv(c1, c2): ...
def legpow(c, pow, maxpower=...): ...
def legder(c, m=..., scl=..., axis=...): ...
def legint(c, m=..., k = ..., lbnd=..., scl=..., axis=...): ...
def legval(x, c, tensor=...): ...
def legval2d(x, y, c): ...
def leggrid2d(x, y, c): ...
def legval3d(x, y, z, c): ...
def leggrid3d(x, y, z, c): ...
def legvander(x, deg): ...
def legvander2d(x, y, deg): ...
def legvander3d(x, y, z, deg): ...
def legfit(x, y, deg, rcond=..., full=..., w=...): ...
def legcompanion(c): ...
def legroots(c): ...
def leggauss(deg): ...
def legweight(x): ...

class Legendre(ABCPolyBase):
    domain: Any
    window: Any
    basis_name: Any



from collections.abc import Callable
from typing import Any

import numpy as np
import numpy.typing as npt

AR: npt.NDArray[np.float64]
func1: Callable[[Any], str]
func2: Callable[[np.integer[Any]], str]

np.array2string(AR, style=None)  # E: Unexpected keyword argument
np.array2string(AR, legacy="1.14")  # E: incompatible type
np.array2string(AR, sign="*")  # E: incompatible type
np.array2string(AR, floatmode="default")  # E: incompatible type
np.array2string(AR, formatter={"A": func1})  # E: incompatible type
np.array2string(AR, formatter={"float": func2})  # E: Incompatible types



import sys
from pathlib import Path
from typing import IO, Any

import numpy as np

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

path1: Path
path2: str

d1 = np.lib.npyio.DataSource(path1)
d2 = np.lib.npyio.DataSource(path2)
d3 = np.lib.npyio.DataSource(None)

assert_type(d1.abspath("..."), str)
assert_type(d2.abspath("..."), str)
assert_type(d3.abspath("..."), str)

assert_type(d1.exists("..."), bool)
assert_type(d2.exists("..."), bool)
assert_type(d3.exists("..."), bool)

assert_type(d1.open("...", "r"), IO[Any])
assert_type(d2.open("...", encoding="utf8"), IO[Any])
assert_type(d3.open("...", newline="/n"), IO[Any])



"""
Tests for miscellaneous (non-magic) ``np.ndarray``/``np.generic`` methods.

More extensive tests are performed for the methods'
function-based counterpart in `../from_numeric.py`.

"""

import sys
import operator
import ctypes as ct
from typing import Any, Literal

import numpy as np
import numpy.typing as npt

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

class SubClass(npt.NDArray[np.object_]): ...

f8: np.float64
B: SubClass
AR_f8: npt.NDArray[np.float64]
AR_i8: npt.NDArray[np.int64]
AR_U: npt.NDArray[np.str_]
AR_V: npt.NDArray[np.void]

ctypes_obj = AR_f8.ctypes

assert_type(AR_f8.__dlpack__(), Any)
assert_type(AR_f8.__dlpack_device__(), tuple[int, Literal[0]])

assert_type(ctypes_obj.data, int)
assert_type(ctypes_obj.shape, ct.Array[np.ctypeslib.c_intp])
assert_type(ctypes_obj.strides, ct.Array[np.ctypeslib.c_intp])
assert_type(ctypes_obj._as_parameter_, ct.c_void_p)

assert_type(ctypes_obj.data_as(ct.c_void_p), ct.c_void_p)
assert_type(ctypes_obj.shape_as(ct.c_longlong), ct.Array[ct.c_longlong])
assert_type(ctypes_obj.strides_as(ct.c_ubyte), ct.Array[ct.c_ubyte])

assert_type(f8.all(), np.bool)
assert_type(AR_f8.all(), np.bool)
assert_type(AR_f8.all(axis=0), Any)
assert_type(AR_f8.all(keepdims=True), Any)
assert_type(AR_f8.all(out=B), SubClass)

assert_type(f8.any(), np.bool)
assert_type(AR_f8.any(), np.bool)
assert_type(AR_f8.any(axis=0), Any)
assert_type(AR_f8.any(keepdims=True), Any)
assert_type(AR_f8.any(out=B), SubClass)

assert_type(f8.argmax(), np.intp)
assert_type(AR_f8.argmax(), np.intp)
assert_type(AR_f8.argmax(axis=0), Any)
assert_type(AR_f8.argmax(out=B), SubClass)

assert_type(f8.argmin(), np.intp)
assert_type(AR_f8.argmin(), np.intp)
assert_type(AR_f8.argmin(axis=0), Any)
assert_type(AR_f8.argmin(out=B), SubClass)

assert_type(f8.argsort(), npt.NDArray[Any])
assert_type(AR_f8.argsort(), npt.NDArray[Any])

assert_type(f8.astype(np.int64).choose([()]), npt.NDArray[Any])
assert_type(AR_f8.choose([0]), npt.NDArray[Any])
assert_type(AR_f8.choose([0], out=B), SubClass)

assert_type(f8.clip(1), npt.NDArray[Any])
assert_type(AR_f8.clip(1), npt.NDArray[Any])
assert_type(AR_f8.clip(None, 1), npt.NDArray[Any])
assert_type(AR_f8.clip(1, out=B), SubClass)
assert_type(AR_f8.clip(None, 1, out=B), SubClass)

assert_type(f8.compress([0]), npt.NDArray[Any])
assert_type(AR_f8.compress([0]), npt.NDArray[Any])
assert_type(AR_f8.compress([0], out=B), SubClass)

assert_type(f8.conj(), np.float64)
assert_type(AR_f8.conj(), npt.NDArray[np.float64])
assert_type(B.conj(), SubClass)

assert_type(f8.conjugate(), np.float64)
assert_type(AR_f8.conjugate(), npt.NDArray[np.float64])
assert_type(B.conjugate(), SubClass)

assert_type(f8.cumprod(), npt.NDArray[Any])
assert_type(AR_f8.cumprod(), npt.NDArray[Any])
assert_type(AR_f8.cumprod(out=B), SubClass)

assert_type(f8.cumsum(), npt.NDArray[Any])
assert_type(AR_f8.cumsum(), npt.NDArray[Any])
assert_type(AR_f8.cumsum(out=B), SubClass)

assert_type(f8.max(), Any)
assert_type(AR_f8.max(), Any)
assert_type(AR_f8.max(axis=0), Any)
assert_type(AR_f8.max(keepdims=True), Any)
assert_type(AR_f8.max(out=B), SubClass)

assert_type(f8.mean(), Any)
assert_type(AR_f8.mean(), Any)
assert_type(AR_f8.mean(axis=0), Any)
assert_type(AR_f8.mean(keepdims=True), Any)
assert_type(AR_f8.mean(out=B), SubClass)

assert_type(f8.min(), Any)
assert_type(AR_f8.min(), Any)
assert_type(AR_f8.min(axis=0), Any)
assert_type(AR_f8.min(keepdims=True), Any)
assert_type(AR_f8.min(out=B), SubClass)

assert_type(f8.prod(), Any)
assert_type(AR_f8.prod(), Any)
assert_type(AR_f8.prod(axis=0), Any)
assert_type(AR_f8.prod(keepdims=True), Any)
assert_type(AR_f8.prod(out=B), SubClass)

assert_type(f8.round(), np.float64)
assert_type(AR_f8.round(), npt.NDArray[np.float64])
assert_type(AR_f8.round(out=B), SubClass)

assert_type(f8.repeat(1), npt.NDArray[np.float64])
assert_type(AR_f8.repeat(1), npt.NDArray[np.float64])
assert_type(B.repeat(1), npt.NDArray[np.object_])

assert_type(f8.std(), Any)
assert_type(AR_f8.std(), Any)
assert_type(AR_f8.std(axis=0), Any)
assert_type(AR_f8.std(keepdims=True), Any)
assert_type(AR_f8.std(out=B), SubClass)

assert_type(f8.sum(), Any)
assert_type(AR_f8.sum(), Any)
assert_type(AR_f8.sum(axis=0), Any)
assert_type(AR_f8.sum(keepdims=True), Any)
assert_type(AR_f8.sum(out=B), SubClass)

assert_type(f8.take(0), np.float64)
assert_type(AR_f8.take(0), np.float64)
assert_type(AR_f8.take([0]), npt.NDArray[np.float64])
assert_type(AR_f8.take(0, out=B), SubClass)
assert_type(AR_f8.take([0], out=B), SubClass)

assert_type(f8.var(), Any)
assert_type(AR_f8.var(), Any)
assert_type(AR_f8.var(axis=0), Any)
assert_type(AR_f8.var(keepdims=True), Any)
assert_type(AR_f8.var(out=B), SubClass)

assert_type(AR_f8.argpartition([0]), npt.NDArray[np.intp])

assert_type(AR_f8.diagonal(), npt.NDArray[np.float64])

assert_type(AR_f8.dot(1), npt.NDArray[Any])
assert_type(AR_f8.dot([1]), Any)
assert_type(AR_f8.dot(1, out=B), SubClass)

assert_type(AR_f8.nonzero(), tuple[npt.NDArray[np.intp], ...])

assert_type(AR_f8.searchsorted(1), np.intp)
assert_type(AR_f8.searchsorted([1]), npt.NDArray[np.intp])

assert_type(AR_f8.trace(), Any)
assert_type(AR_f8.trace(out=B), SubClass)

assert_type(AR_f8.item(), float)
assert_type(AR_U.item(), str)

assert_type(AR_f8.ravel(), npt.NDArray[np.float64])
assert_type(AR_U.ravel(), npt.NDArray[np.str_])

assert_type(AR_f8.flatten(), npt.NDArray[np.float64])
assert_type(AR_U.flatten(), npt.NDArray[np.str_])

assert_type(AR_f8.reshape(1), npt.NDArray[np.float64])
assert_type(AR_U.reshape(1), npt.NDArray[np.str_])

assert_type(int(AR_f8), int)
assert_type(int(AR_U), int)

assert_type(float(AR_f8), float)
assert_type(float(AR_U), float)

assert_type(complex(AR_f8), complex)

assert_type(operator.index(AR_i8), int)

assert_type(AR_f8.__array_wrap__(B), npt.NDArray[np.object_])

assert_type(AR_V[0], Any)
assert_type(AR_V[0, 0], Any)
assert_type(AR_V[AR_i8], npt.NDArray[np.void])
assert_type(AR_V[AR_i8, AR_i8], npt.NDArray[np.void])
assert_type(AR_V[AR_i8, None], npt.NDArray[np.void])
assert_type(AR_V[0, ...], npt.NDArray[np.void])
assert_type(AR_V[[0]], npt.NDArray[np.void])
assert_type(AR_V[[0], [0]], npt.NDArray[np.void])
assert_type(AR_V[:], npt.NDArray[np.void])
assert_type(AR_V["a"], npt.NDArray[Any])
assert_type(AR_V[["a", "b"]], npt.NDArray[np.void])

assert_type(AR_f8.dump("test_file"), None)
assert_type(AR_f8.dump(b"test_file"), None)
with open("test_file", "wb") as f:
    assert_type(AR_f8.dump(f), None)

assert_type(AR_f8.__array_finalize__(None), None)
assert_type(AR_f8.__array_finalize__(B), None)
assert_type(AR_f8.__array_finalize__(AR_f8), None)



import numpy as np
import numpy.typing as npt

AR_U: npt.NDArray[np.str_]

def func() -> bool: ...

np.testing.assert_(True, msg=1)  # E: incompatible type
np.testing.build_err_msg(1, "test")  # E: incompatible type
np.testing.assert_almost_equal(AR_U, AR_U)  # E: incompatible type
np.testing.assert_approx_equal([1, 2, 3], [1, 2, 3])  # E: incompatible type
np.testing.assert_array_almost_equal(AR_U, AR_U)  # E: incompatible type
np.testing.assert_array_less(AR_U, AR_U)  # E: incompatible type
np.testing.assert_string_equal(b"a", b"a")  # E: incompatible type

np.testing.assert_raises(expected_exception=TypeError, callable=func)  # E: No overload variant
np.testing.assert_raises_regex(expected_exception=TypeError, expected_regex="T", callable=func)  # E: No overload variant

np.testing.assert_allclose(AR_U, AR_U)  # E: incompatible type
np.testing.assert_array_almost_equal_nulp(AR_U, AR_U)  # E: incompatible type
np.testing.assert_array_max_ulp(AR_U, AR_U)  # E: incompatible type

np.testing.assert_warns(warning_class=RuntimeWarning, func=func)  # E: No overload variant
np.testing.assert_no_warnings(func=func)  # E: No overload variant
np.testing.assert_no_warnings(func, None)  # E: Too many arguments
np.testing.assert_no_warnings(func, test=None)  # E: Unexpected keyword argument

np.testing.assert_no_gc_cycles(func=func)  # E: No overload variant



import sys

import numpy as np
from numpy._typing import _80Bit, _96Bit, _128Bit, _256Bit

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

assert_type(np.uint128(), np.unsignedinteger[_128Bit])
assert_type(np.uint256(), np.unsignedinteger[_256Bit])

assert_type(np.int128(), np.signedinteger[_128Bit])
assert_type(np.int256(), np.signedinteger[_256Bit])

assert_type(np.float80(), np.floating[_80Bit])
assert_type(np.float96(), np.floating[_96Bit])
assert_type(np.float128(), np.floating[_128Bit])
assert_type(np.float256(), np.floating[_256Bit])

assert_type(np.complex160(), np.complexfloating[_80Bit, _80Bit])
assert_type(np.complex192(), np.complexfloating[_96Bit, _96Bit])
assert_type(np.complex256(), np.complexfloating[_128Bit, _128Bit])
assert_type(np.complex512(), np.complexfloating[_256Bit, _256Bit])



from pandas import DataFrame as _pd_DataFrame
from matplotlib import pyplot as plt
from matplotlib.animation import (FuncAnimation as _plt_FuncAnimation, PillowWriter as _plt_PillowWriter)
from matplotlib.figure import Figure as _plt_Figure
from matplotlib.axes._axes import Axes as _plt_Axes
from .plot_utils import (
    create_grid_cell_patches, 
    create_circle_patches
)

## Animation / GIF for varying grid spacings

def set_animation_frames( 
    relevantGridSizes:_pd_DataFrame,
    allGridSizes:_pd_DataFrame,
    largest:float=400,
    frames_between_relevant:int = 6
    )->list:
    """
    Returns vector of gridsizes -> each grid_spacing will be an iteration frame
    """
    # define GIF steps
    animation_frames = [largest]
    allGridSizes_indexes = list(allGridSizes.index)

    for gridsize_0,gridsize_1 in zip(relevantGridSizes.index[:-1], relevantGridSizes.index[1:]):
        
        pos_0, pos_1 = allGridSizes_indexes.index(gridsize_0), allGridSizes_indexes.index(gridsize_1)
        
        #
        if (pos_1-pos_0) > frames_between_relevant:
            # number of position that should moved along per allGridSizes_indexes frame 
            stepsize_factor = (pos_1-pos_0)//frames_between_relevant
            # move an aditional positon for the n<stepsize_adjust frames
            stepsize_adjust = (pos_1-pos_0)%frames_between_relevant
            
            animation_frames += [
                allGridSizes_indexes[pos_0+s*stepsize_factor+int(stepsize_adjust<s)] 
                for s in range(frames_between_relevant)
                ]
        else:
            #
            animation_frames += allGridSizes_indexes[pos_0+1:pos_1+1]
            
    print('Animate GIF with '+str(len(animation_frames))+' frames.')

    return animation_frames


def add_coverage_plot(
        fig:_plt_Figure,
        allGridSizes:_pd_DataFrame,
        ymax:float=3,
        yticks:list=[.5,1]
    ) -> _plt_Axes:
    """
    adds new axis to fig and create a stackplot on it that shows:
     - percent of area within grid cells that are completly within
     - stacked percent of area within grid cells that are potetially intersected
     - 
    Returns: axes 
    """
    ax_coverage = fig.add_axes([0.6, 0.63, 0.3, 0.25])
        
    ax_coverage.stackplot(
        allGridSizes['grid_spacing'],allGridSizes['share_contain']*100,allGridSizes['share_overlap']*100,
        colors=['green', 'red']
        )
    ax_coverage.plot(allGridSizes['grid_spacing'], allGridSizes['share_overlap']*100, color='#000')
    min_x, max_x = min(allGridSizes['grid_spacing']),max(allGridSizes['grid_spacing'])
    ax_coverage.set_xlim(min_x,max_x)
    ax_coverage.set_ylim(0,100*ymax)
    
    ax_coverage.set_yticks([100*x for x in sorted(yticks+[round(ymax)])])
    
    # add hline at 100%
    ax_coverage.hlines(y=100,xmin=min_x,xmax=max_x,linewidth=2,color='#000')
    
    return ax_coverage
#

def add_cell_counter_plot(
        fig:_plt_Figure,
        allGridSizes:_pd_DataFrame,
        ymax:float=3,
        yticks:list=[.5,1]
    ) -> _plt_Axes:
    """
    adds new axis to fig and create a step plot on that counts number of cells:
    - which are fully contained s.t. the sum of emplyoment will be requested
    - which (might be) intersected s.t. all points within will need to be checked whether they are within r  
    Returns: axes 
    """
    ax_counter = fig.add_axes([0.6, 0.43, 0.3, 0.25])
        
    ax_counter.plot(
        allGridSizes['grid_spacing'], allGridSizes['contain_count'], color='green'
        )
    ax_counter.plot(
        allGridSizes['grid_spacing'], allGridSizes['overlap_count'], color='red'
        )
    ax_counter.plot(allGridSizes['grid_spacing'], allGridSizes['overlap_count'], color='#000')
    minx, maxx = min(allGridSizes['grid_spacing']),max(allGridSizes['grid_spacing'])
    ax_counter.set_xlim(minx,maxx)
    ax_counter.set_ylim(0,max(allGridSizes['contain_count']+allGridSizes['overlap_count']))
    
    ax_counter.set_yticks(yticks)
    
    return ax_counter
#

def update_fig_title(fig, grid_spacing:float,it:dict):
    """
    updates fig.suptitle with rounded grid_spacing of current frame
    """
    # display grid_spacing as title
    gridsize_rnd = round(grid_spacing,2) 
    fig.suptitle(
        (' ' if gridsize_rnd >= 100 else '') +
        str(gridsize_rnd)[:5+int(gridsize_rnd >= 100)] +
        '0'*max(0,len(str(gridsize_rnd))-int(gridsize_rnd >= 100)-3) +
        '. Frame: '+str(it)
    )
#

def update_main_axis_frame(
        allGridSizes:_pd_DataFrame, 
        grid_spacing:float, 
        ax_main,
        ax_min:float=0, 
        ax_max:float=1,
        r:float=750,
    ) -> list:
    # clear previous picture
    ax_main.clear()
    
    listOfPatches = []
    # create patches for grid cell: green=overlapped, red= (potentially) intersected, grey=outside 
    listOfPatches += create_grid_cell_patches( 
        grid_spacing=grid_spacing, 
        ax_min=ax_min, 
        ax_max=ax_max,
        contain_cells_row_col=allGridSizes.loc[grid_spacing, 'contain_ids'],
        overlap_cells_row_col=allGridSizes.loc[grid_spacing, 'overlap_ids'],
        )
    
    # create patches for circle around bottom left and top right corner of center grid cell
    listOfPatches += create_circle_patches(
        grid_spacing=grid_spacing,
        r=r
    )

    for patchToAdd in listOfPatches:
        ax_main.add_patch(patchToAdd)

    points = [p[0] for p in [
        ax_main.plot(0, 0, marker='x', color='black'),
    ]]

    # set tickmarks and limits
    ax_main.set_yticks([r-grid_spacing/2, r, r+grid_spacing/2]) 
    ax_main.set_xticks([r-grid_spacing/2, r, r+grid_spacing/2]) 
    ax_main.set_xlim(ax_min,ax_max)
    ax_main.set_ylim(ax_min,ax_max)
    ax_main.set_aspect('equal', adjustable='box')

    return *listOfPatches, *points, #, point1,
#

def update_coverage_axis_frame(
    grid_spacing:float,
    max_gridsize:float,
    min_gridsize:float,
    ax_coverage,
    vline,
    ) -> tuple:
    """
    
    """
    # update vline
    if vline != None:
            vline.remove()
    vline = ax_coverage.vlines(x=grid_spacing, ymin=0, ymax=3)
    ax_coverage.set_xticks([round(x,0) for x in set([min_gridsize, grid_spacing, max_gridsize])])
    return (vline,)
#

def create_optimal_grid_spacing_gif (
        relevantGridSizes:_pd_DataFrame,
        allGridSizes:_pd_DataFrame,
        largest:float=400,
        smallest:float=80,
        r:float=750,
        frames_between_relevant:int = 6,
        FuncAnimation_interval:int=500,
        FuncAnimation_repeat:bool=True,
        FuncAnimation_repeat_delay:int=1500,
        FuncAnimation_cache_frame_data:bool=True,
        dpi:int=50,
        PillowWriterFps:int=10,
        output_dir:str='./opt_grid_'
        )->tuple:
    
    print('Create GIF')
    # initialise plot and main axis
    fig,ax_main = plt.subplots()

    # define ax limits for main axis 
    min_gridsize = min(relevantGridSizes['grid_spacing'])
    max_gridsize = max(relevantGridSizes['grid_spacing'])
    max_relevant_cells = max(relevantGridSizes['contain_count']+relevantGridSizes['overlap_count'])
    ax_min=-max_gridsize/2
    ax_max= max_gridsize*(max(relevantGridSizes['cell_steps_max'])-.5)    
    
    # add stackplot to top right corner
    ax_coverage = add_coverage_plot(
        fig=fig, allGridSizes=allGridSizes,
        ymax=1.01*max(relevantGridSizes['share_contain']+relevantGridSizes['share_overlap']),
        yticks=[.5,1])

    # ax_counter = add_cell_counter_plot(
    #     fig=fig, 
    #     allGridSizes=allGridSizes,
    #     ymax=max_relevant_cells,
    #     yticks=[int(x*max_relevant_cells/4) for x in range(1,5)]
    #     )
    
    # 
    an_dct={
        'it':0,
        'vline':None
        }

    def animate(grid_spacing):
        """
        function to pass into FuncAnimation to create GIF. 
        Will be supplied with list of gridsizes to iterate over.
        """
        an_dct['it']=an_dct['it']+1
        
        update_fig_title(fig=fig,grid_spacing=grid_spacing,it=an_dct['it'])
        
        # update main axis:
        artistObjects = update_main_axis_frame(
            allGridSizes=allGridSizes, 
            grid_spacing=grid_spacing,
            ax_main=ax_main, 
            ax_min=ax_min, 
            ax_max=ax_max,
            r=r,
        )
        
        #
        an_dct['vline'], = update_coverage_axis_frame(
            grid_spacing=grid_spacing,
            max_gridsize=max_gridsize,
            min_gridsize=min_gridsize,
            ax_coverage=ax_coverage,
            vline=an_dct['vline']
            ) 

        return artistObjects
    #

    animation_frames = set_animation_frames(
        relevantGridSizes=relevantGridSizes,
        allGridSizes=allGridSizes,
        largest=largest,
        frames_between_relevant=frames_between_relevant
        )

    ani = _plt_FuncAnimation(
        fig, 
        animate, 
        interval=FuncAnimation_interval, 
        blit=True, 
        repeat=FuncAnimation_repeat, repeat_delay=FuncAnimation_repeat_delay, 
        cache_frame_data=FuncAnimation_cache_frame_data,
        frames=animation_frames
        )    

    print('Save GIF')
    ani.save(
        output_dir+str(largest)+"_"+str(smallest)+".gif", 
        dpi=dpi, 
        writer=_plt_PillowWriter(fps=PillowWriterFps)
        )

    return fig,ax_main,ax_coverage,#ax_counter,
#


from numpy.distutils.fcompiler import FCompiler

compilers = ['NVHPCFCompiler']

class NVHPCFCompiler(FCompiler):
    """ NVIDIA High Performance Computing (HPC) SDK Fortran Compiler
   
    https://developer.nvidia.com/hpc-sdk
   
    Since august 2020 the NVIDIA HPC SDK includes the compilers formerly known as The Portland Group compilers,
    https://www.pgroup.com/index.htm.
    See also `numpy.distutils.fcompiler.pg`.
    """

    compiler_type = 'nv'
    description = 'NVIDIA HPC SDK'
    version_pattern = r'\s*(nvfortran|.+ \(aka nvfortran\)) (?P<version>[\d.-]+).*'

    executables = {
        'version_cmd': ["<F90>", "-V"],
        'compiler_f77': ["nvfortran"],
        'compiler_fix': ["nvfortran", "-Mfixed"],
        'compiler_f90': ["nvfortran"],
        'linker_so': ["<F90>"],
        'archiver': ["ar", "-cr"],
        'ranlib': ["ranlib"]
    }
    pic_flags = ['-fpic']

    module_dir_switch = '-module '
    module_include_switch = '-I'

    def get_flags(self):
        opt = ['-Minform=inform', '-Mnosecond_underscore']
        return self.pic_flags + opt

    def get_flags_opt(self):
        return ['-fast']

    def get_flags_debug(self):
        return ['-g']

    def get_flags_linker_so(self):
        return ["-shared", '-fpic']

    def runtime_library_dir_option(self, dir):
        return '-R%s' % dir

if __name__ == '__main__':
    from distutils import log
    log.set_verbosity(2)
    from numpy.distutils import customized_fcompiler
    print(customized_fcompiler(compiler='nv').get_version())



import os
from collections.abc import Sequence, Iterable
from typing import (
    Any,
    TypeVar,
    overload,
    Protocol,
    SupportsIndex,
    Literal
)

from numpy import (
    ndarray,
    dtype,
    generic,
    void,
    _ByteOrder,
    _SupportsBuffer,
    _ShapeType,
    _DType_co,
    _OrderKACF,
)

from numpy._typing import (
    ArrayLike,
    DTypeLike,
    NDArray,
    _ShapeLike,
    _ArrayLikeInt_co,
    _ArrayLikeVoid_co,
    _NestedSequence,
)

_SCT = TypeVar("_SCT", bound=generic)

_RecArray = recarray[Any, dtype[_SCT]]

class _SupportsReadInto(Protocol):
    def seek(self, offset: int, whence: int, /) -> object: ...
    def tell(self, /) -> int: ...
    def readinto(self, buffer: memoryview, /) -> int: ...

class record(void):
    def __getattribute__(self, attr: str) -> Any: ...
    def __setattr__(self, attr: str, val: ArrayLike) -> None: ...
    def pprint(self) -> str: ...
    @overload
    def __getitem__(self, key: str | SupportsIndex) -> Any: ...
    @overload
    def __getitem__(self, key: list[str]) -> record: ...

class recarray(ndarray[_ShapeType, _DType_co]):
    # NOTE: While not strictly mandatory, we're demanding here that arguments
    # for the `format_parser`- and `dtype`-based dtype constructors are
    # mutually exclusive
    @overload
    def __new__(
        subtype,
        shape: _ShapeLike,
        dtype: None = ...,
        buf: None | _SupportsBuffer = ...,
        offset: SupportsIndex = ...,
        strides: None | _ShapeLike = ...,
        *,
        formats: DTypeLike,
        names: None | str | Sequence[str] = ...,
        titles: None | str | Sequence[str] = ...,
        byteorder: None | _ByteOrder = ...,
        aligned: bool = ...,
        order: _OrderKACF = ...,
    ) -> recarray[Any, dtype[record]]: ...
    @overload
    def __new__(
        subtype,
        shape: _ShapeLike,
        dtype: DTypeLike,
        buf: None | _SupportsBuffer = ...,
        offset: SupportsIndex = ...,
        strides: None | _ShapeLike = ...,
        formats: None = ...,
        names: None = ...,
        titles: None = ...,
        byteorder: None = ...,
        aligned: Literal[False] = ...,
        order: _OrderKACF = ...,
    ) -> recarray[Any, dtype[Any]]: ...
    def __array_finalize__(self, obj: object) -> None: ...
    def __getattribute__(self, attr: str) -> Any: ...
    def __setattr__(self, attr: str, val: ArrayLike) -> None: ...
    @overload
    def __getitem__(self, indx: (
        SupportsIndex
        | _ArrayLikeInt_co
        | tuple[SupportsIndex | _ArrayLikeInt_co, ...]
    )) -> Any: ...
    @overload
    def __getitem__(self: recarray[Any, dtype[void]], indx: (
        None
        | slice
        | ellipsis
        | SupportsIndex
        | _ArrayLikeInt_co
        | tuple[None | slice | ellipsis | _ArrayLikeInt_co | SupportsIndex, ...]
    )) -> recarray[Any, _DType_co]: ...
    @overload
    def __getitem__(self, indx: (
        None
        | slice
        | ellipsis
        | SupportsIndex
        | _ArrayLikeInt_co
        | tuple[None | slice | ellipsis | _ArrayLikeInt_co | SupportsIndex, ...]
    )) -> ndarray[Any, _DType_co]: ...
    @overload
    def __getitem__(self, indx: str) -> NDArray[Any]: ...
    @overload
    def __getitem__(self, indx: list[str]) -> recarray[_ShapeType, dtype[record]]: ...
    @overload
    def field(self, attr: int | str, val: None = ...) -> Any: ...
    @overload
    def field(self, attr: int | str, val: ArrayLike) -> None: ...

class format_parser:
    dtype: dtype[void]
    def __init__(
        self,
        formats: DTypeLike,
        names: None | str | Sequence[str],
        titles: None | str | Sequence[str],
        aligned: bool = ...,
        byteorder: None | _ByteOrder = ...,
    ) -> None: ...

__all__: list[str]

@overload
def fromarrays(
    arrayList: Iterable[ArrayLike],
    dtype: DTypeLike = ...,
    shape: None | _ShapeLike = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
) -> _RecArray[Any]: ...
@overload
def fromarrays(
    arrayList: Iterable[ArrayLike],
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    *,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
) -> _RecArray[record]: ...

@overload
def fromrecords(
    recList: _ArrayLikeVoid_co | tuple[Any, ...] | _NestedSequence[tuple[Any, ...]],
    dtype: DTypeLike = ...,
    shape: None | _ShapeLike = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
) -> _RecArray[record]: ...
@overload
def fromrecords(
    recList: _ArrayLikeVoid_co | tuple[Any, ...] | _NestedSequence[tuple[Any, ...]],
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    *,
    formats: DTypeLike = ...,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
) -> _RecArray[record]: ...

@overload
def fromstring(
    datastring: _SupportsBuffer,
    dtype: DTypeLike,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
) -> _RecArray[record]: ...
@overload
def fromstring(
    datastring: _SupportsBuffer,
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    *,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
) -> _RecArray[record]: ...

@overload
def fromfile(
    fd: str | bytes | os.PathLike[str] | os.PathLike[bytes] | _SupportsReadInto,
    dtype: DTypeLike,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
) -> _RecArray[Any]: ...
@overload
def fromfile(
    fd: str | bytes | os.PathLike[str] | os.PathLike[bytes] | _SupportsReadInto,
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    *,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
) -> _RecArray[record]: ...

@overload
def array(
    obj: _SCT | NDArray[_SCT],
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
    copy: bool = ...,
) -> _RecArray[_SCT]: ...
@overload
def array(
    obj: ArrayLike,
    dtype: DTypeLike,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
    copy: bool = ...,
) -> _RecArray[Any]: ...
@overload
def array(
    obj: ArrayLike,
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    *,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
    copy: bool = ...,
) -> _RecArray[record]: ...
@overload
def array(
    obj: None,
    dtype: DTypeLike,
    shape: _ShapeLike,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
    copy: bool = ...,
) -> _RecArray[Any]: ...
@overload
def array(
    obj: None,
    dtype: None = ...,
    *,
    shape: _ShapeLike,
    offset: int = ...,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
    copy: bool = ...,
) -> _RecArray[record]: ...
@overload
def array(
    obj: _SupportsReadInto,
    dtype: DTypeLike,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    formats: None = ...,
    names: None = ...,
    titles: None = ...,
    aligned: bool = ...,
    byteorder: None = ...,
    copy: bool = ...,
) -> _RecArray[Any]: ...
@overload
def array(
    obj: _SupportsReadInto,
    dtype: None = ...,
    shape: None | _ShapeLike = ...,
    offset: int = ...,
    *,
    formats: DTypeLike,
    names: None | str | Sequence[str] = ...,
    titles: None | str | Sequence[str] = ...,
    aligned: bool = ...,
    byteorder: None | _ByteOrder = ...,
    copy: bool = ...,
) -> _RecArray[record]: ...



# intersection of two circles with same radius
from numpy import (linspace as _np_linspace, array as _np_array, sign as _np_sign)
from math import sin as _math_sin, cos as _math_cos, atan2 as _math_atan2, pi as _math_pi, acos as _math_acos , sin as _math_asin, log10 as _math_log10
from ..utils.general import angle, angles_to_origin, angle_to



def enusure_zero_type(nums:tuple, tp:type=float):
    return type(nums)([num if num != 0 else tp(num) for num in nums])

def circle_line_segment_intersection(circle_center, circle_radius, pt1, pt2, full_line=True, precision:int=13):
    """ Find the points at which a circle intersects a line-segment.  This can happen at 0, 1, or 2 points.

    :param circle_center: The (x, y) location of the circle center
    :param circle_radius: The radius of the circle
    :param pt1: The (x, y) location of the first point of the segment
    :param pt2: The (x, y) location of the second point of the segment
    :param full_line: True to find intersections along full line - not just in the segment.  False will just return intersections within the segment.
    :param tangent_tol: Numerical tolerance at which we decide the intersections are close enough to consider it a tangent
    :return Sequence[Tuple[float, float]]: A list of length 0, 1, or 2, where each element is a point at which the circle intercepts a line segment.

    Note: We follow: http://mathworld.wolfram.com/Circle-LineIntersection.html
    TAKEN FROM: https://stackoverflow.com/a/59582674
    """
    tangent_tol = 10**-(precision)
    (p1x, p1y), (p2x, p2y), (cx, cy) = pt1, pt2, circle_center
    (x1, y1), (x2, y2) = (p1x - cx, p1y - cy), (p2x - cx, p2y - cy)
    dx, dy = (x2 - x1), (y2 - y1)
    dr = (dx ** 2 + dy ** 2)**.5
    big_d = (x1 * y2 - x2 * y1)
    discriminant = (circle_radius ** 2 * dr ** 2 - big_d ** 2)

    if discriminant < 0:  # No intersection between circle and line
        return []
    else:  # There may be 0, 1, or 2 intersections with the segment
        intersections = [
            (cx + (big_d * dy + sign * (-1 if dy < 0 else 1) * dx * discriminant**.5) / dr ** 2,
             cy + (-big_d * dx + sign * abs(dy) * discriminant**.5) / dr ** 2)
            for sign in ((1, -1) if dy < 0 else (-1, 1))]  # This makes sure the order along the segment is correct
        if not full_line:  # If only considering the segment, filter out intersections that do not fall within the segment
            fraction_along_segment = [(xi - p1x) / dx if abs(dx) > abs(dy) else (yi - p1y) / dy for xi, yi in intersections]
            intersections = [pt for pt, frac in zip(intersections, fraction_along_segment) if 0 <= frac <= 1]
        intersections = [enusure_zero_type((round(x, precision), round(y, precision)),float) for x,y in intersections]
        if len(intersections) == 2 and abs(discriminant) <= tangent_tol:  # If line is tangent to circle, return just one point (as both intersections have same location)
            return [intersections[0]]
        else:
            return intersections
#

def isBetween(a, b, c, precision:int=13):
    crossproduct = (c[1] - a[1]) * (b[0] - a[0]) - (c[0] - a[0]) * (b[1] - a[1])

    # compare versus epsilon for floating point values, or != 0 if using integers
    if abs(crossproduct) > 10**-precision:
        return False

    dotproduct = (c[0] - a[0]) * (b[0] - a[0]) + (c[1] - a[1])*(b[1] - a[1])
    if dotproduct < 0:
        return False

    squaredlengthba = (b[0] - a[0])*(b[0] - a[0]) + (b[1] - a[1])*(b[1] - a[1])
    if dotproduct > squaredlengthba:
        return False

    return True
#

def det(a, b):
    return a[0] * b[1] - a[1] * b[0]
#

def line_intersection(line1, line2, precision:int=13):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    div = det(xdiff, ydiff)
    if div == 0:
       return []
    d = (det(*line1), det(*line2))
    x,y = (det(d, xdiff) / div, det(d, ydiff) / div)
    if x==0:
        x = 0.
    if y==0:
        y = 0.
    if not isBetween(line1[0], line1[1], (x,y), precision=precision):
        return []
    
    return [(x,y)]
#

def filter_pts_on_arc(
        pts,
        arc_center,
        arc_angle_min:float=0,
        arc_angle_max:float=360,
        precision:int=13,
):
    """
    Filter points on circle to those within angle_min and angle_max 
    """
    return [
        pt for pt, ngl in zip(pts, angles_to_origin(pts, arc_center)) 
        if arc_angle_min - 1*10**-(precision + 0) <= ngl and ngl <= arc_angle_max + 1*10**-(precision + 0)
    ]
#

def  intersections_pts_two_circles_same_radius(
        center_1,
        center_2,
        r:float,
        precision:int=13
        ):
    """
    Returns list intersections points of two circle with same radius
    """
    circle_1_x, circle_1_y = center_1
    circle_2_x, circle_2_y = center_2

    dist = ((circle_1_x-circle_2_x)**2 + (circle_1_y-circle_2_y)**2)**.5
    if dist > 2*r:
        return []
    if dist == 2*r:
        return [(
            round((circle_1_x + circle_2_x) / 2, precision), 
            round((circle_1_y + circle_2_y) / 2, precision),
            )]

    alpha = _math_acos(dist/2/r)
    slope_angle = angle(circle_1_x, circle_1_y, circle_2_x, circle_2_y)
    angle_itx1 = alpha-slope_angle
    angle_itx2 = _math_pi*2-slope_angle-alpha
    itx_coords = [
        (
            round(r*_math_cos(angle_itx1) + circle_1_x, precision),
            round(r*_math_sin(angle_itx1)+ circle_1_y, precision),
        ), (
            round(r*_math_cos(angle_itx2) + circle_1_x, precision),
            round(r*_math_sin(angle_itx2) + circle_1_y, precision),
        )
    ]
    itx_coords = [tuple([e if e != 0 else 0. for e in el]) for el in itx_coords] # ensure -0.0 is +0.0
    
    return itx_coords
# 

def intersections_pts_arc_to_circle(
        circle_center,
        arc_center,
        r:float,
        arc_angle_min:float=0,
        arc_angle_max:float=360,
        precision:int=10,
        ):
    """
    Get intersections points of arc and circle (filtering out points that are not on arc)
    """
    # pts = intersections_pts_two_circles_same_radius(
    #         center_1=circle_center,
    #         center_2=arc_center,
    #         r=r,
    #     )
    # filtered_pts = filter_pts_on_arc(
    #     pts=pts,
    #     arc_center=arc_center,
    #     arc_angle_min=arc_angle_min,
    #     arc_angle_max=arc_angle_max,
    #     precision=precision,
    # )
    # if len(pts) != len(filtered_pts):
    #     pass
    #     print("FILTER",len(pts), len(filtered_pts), (arc_angle_min, arc_angle_max, ), pts)
    
    return filter_pts_on_arc(
        pts=intersections_pts_two_circles_same_radius(
            center_1=circle_center,
            center_2=arc_center,
            r=r,
            precision=13,
        ),
        arc_center=arc_center,
        arc_angle_min=arc_angle_min,
        arc_angle_max=arc_angle_max,
        precision=precision,
    )
#


from distutils.unixccompiler import UnixCCompiler

class FujitsuCCompiler(UnixCCompiler):

    """
    Fujitsu compiler.
    """

    compiler_type = 'fujitsu'
    cc_exe = 'fcc'
    cxx_exe = 'FCC'

    def __init__(self, verbose=0, dry_run=0, force=0):
        UnixCCompiler.__init__(self, verbose, dry_run, force)
        cc_compiler = self.cc_exe
        cxx_compiler = self.cxx_exe
        self.set_executables(
            compiler=cc_compiler +
            ' -O3 -Nclang -fPIC',
            compiler_so=cc_compiler +
            ' -O3 -Nclang -fPIC',
            compiler_cxx=cxx_compiler +
            ' -O3 -Nclang -fPIC',
            linker_exe=cc_compiler +
            ' -lfj90i -lfj90f -lfjsrcinfo -lelf -shared',
            linker_so=cc_compiler +
            ' -lfj90i -lfj90f -lfjsrcinfo -lelf -shared'
            )



from numpy import (
    array as _np_array, 
    linspace as _np_linspace,
    stack as _np_stack,
    arange as _np_arange, 
    unique as _np_unique,
    zeros as _np_zeros,
)
from pyproj import Transformer
from pandas import DataFrame as _pd_DataFrame
from math import log10 as _math_log10#,ceil,asin,acos
from aabpl.utils.general import flatten_list, arr_to_tpls
from aabpl.illustrations.plot_utils import map_2D_to_rgb, get_2D_rgb_colobar_kwargs
from .radius_search_class import (
    aggregate_point_data_to_cells,
    aggreagate_point_data_to_disks_vectorized
)
from .pts_to_offset_regions import assign_points_to_cell_regions
from aabpl.valid_area import disk_cell_intersection_area
from aabpl.testing.test_performance import time_func_perf
import matplotlib.pyplot as plt
from matplotlib.pyplot import get_cmap as _plt_get_cmap
from shapely.geometry import Polygon, Point
from shapely.ops import unary_union
from geopandas import GeoDataFrame as _gpd_GeoDataFrame

class Bounds(object):
    __slots__ = ('xmin', 'xmax', 'ymin', 'ymax', 'np_array_of_bounds') # use this syntax to save some memory. also only create vars that are really neccessary
    def __init__(self, xmin:float, xmax:float, ymin:float, ymax:float):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
    #
#

# TODO potentially add nestes GridCell classtype the 
class GridCell(object):
    # __slots__ = ('id', 'row_col_nr', ...) # use this syntax to save some memory. also only create vars that are really neccessary
    def __init__(self, row_nr, col_nr, y_steps, x_steps):
        self.id = (row_nr,col_nr)
        self.centroid = (y_steps[row_nr:row_nr+2].sum()/2,x_steps[col_nr:col_nr+2].sum()/2)
        self.bounds = Bounds(xmin=x_steps[col_nr], ymin=y_steps[row_nr], xmax=x_steps[col_nr+1], ymax=y_steps[row_nr+1])
        self.pt_ids = []
        self.excluded = None

    def add_pt_id(self, pt_id):
        self.pt_ids = [*self.pt_ids, pt_id]
        
    def add_pt_ids(self, pt_ids):
        self.pt_ids = [*self.pt_ids, *pt_ids]
    
    def set_excluded_area(self,excluded_area):
        pass

    def make_sparse(self):
        # self.pt_ids # make _np_array or tuple
        # tighten bounds
        pass            
    def tighten_bounds(self, pt_ids):
        self.pt_ids = [*self.pt_ids, *pt_ids]
    
    def plot(self, facecolor, edgecolor, ):
        pass
#


class Grid(object):
    """

    """
    @time_func_perf
    def __init__(
        self,
        xmin:float,
        xmax:float,
        ymin:float,
        ymax:float,
        initial_crs:str,
        local_crs:str,
        set_fixed_spacing:float=None,
        r:float=750,
        n_points:int=10000,
        silent = False,
        ):

        """

        """
        if set_fixed_spacing:
            spacing = set_fixed_spacing
        else:
            # find optimal spacing TODO
            print("TODO find optimal spacing for",r, n_points)
            spacing = 1.
        self.spacing = spacing

        # TODO total_bounds should also contain excluded area if not contained 
        # min(points.total_bounds+r, max(points.total_bounds, excluded_area_total_bound))  
        self.initial_crs = initial_crs
        self.local_crs = local_crs
        x_padding = ((xmin-xmax) % spacing)/2
        y_padding = ((ymin-ymax) % spacing)/2
        
        self.n_x_steps = n_x_steps = -int((xmin-xmax)/spacing) # round up
        self.n_y_steps = n_y_steps = -int((ymin-ymax)/spacing) # round up 
        self.total_bounds = total_bounds = Bounds(xmin=xmin-x_padding,xmax=xmax+x_padding,ymin=ymin-y_padding,ymax=ymax+y_padding)
        
        self.x_steps = x_steps = _np_linspace(total_bounds.xmin, total_bounds.xmax, n_x_steps)
        # self.y_steps = y_steps = _np_linspace(total_bounds.ymax, total_bounds.ymin, n_y_steps)
        self.y_steps = y_steps = _np_linspace(total_bounds.ymin, total_bounds.ymax, n_y_steps)
        
        self.id_y_mult = id_y_mult = 10**(int(_math_log10(n_x_steps))+1)
        
        self.row_ids = _np_arange(n_y_steps-1)
        self.col_ids = _np_arange(n_x_steps-1)
        
        self.ids = tuple(flatten_list([[(row_id, col_id) for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]))
        class CellDict(dict):
            def __init__(self, x_steps, y_steps, id_y_mult):
                self.x_steps = x_steps
                self.y_steps = y_steps
                self.id_y_mult = id_y_mult

            def id_to_row_col(self,id:int): 
                return (id // self.id_y_mult, id % self.id_y_mult)
            
            def row_col_to_id(self,row_nr:int,col_nr:int): 
                return row_nr * self.id_y_mult + col_nr
            
            def add_new(self,row,col):
                setattr(self, str(self.row_col_to_id(row,col)), GridCell(
                    row, col, self.x_steps, self.y_steps
                ))
            
            def get_by_row_col(self,row,col):
                return getattr(self, str(self.row_col_to_id(row,col)))
            
            def get_by_id(self,id):
                return getattr(self, str(id))
            
            def get_or_create(self,row,col):
                id = str(self.row_col_to_id(row,col))
                if not hasattr(self, id):
                    self.add_new(row,col)
                return self.get_by_id(id)
            
            def add_pts(self, pts, row, col):
                cell = self.get_or_create(row,col)
                cell.add_pts(pts)
            
            def add_pt(self, pt, row, col):
                cell = self.get_or_create(row,col)
                cell.add_pt(pt)
            
        self.row_col_stack = _np_stack([
                    _np_array([[row_id for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]).flatten(),
                    _np_array([[col_id for col_id in range(n_x_steps-1)] for row_id in range(n_y_steps-1)]).flatten(),
                ])
        self.cells = CellDict(x_steps=x_steps, y_steps=y_steps, id_y_mult=id_y_mult,)

        self.cells = _np_array([[
            GridCell(row_id, col_id, y_steps=y_steps, x_steps=x_steps) for col_id in range(n_x_steps-1)
            ] for row_id in range(n_y_steps-1)]).flatten()

        self.row_col_to_centroid = {g_row_col:centroid for (g_row_col,centroid) in flatten_list([
                [((row_id,col_id),(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean())) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )}
        self.centroids = _np_array([centroid for centroid in flatten_list([
                [(x_steps[col_id:col_id+2].mean(), y_steps[row_id:row_id+2].mean()) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )])
        self.id_to_bounds = {row_id*id_y_mult+col_id: bounds for ((row_id,col_id),bounds) in flatten_list([
                [((row_id,col_id),((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1]))) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )}
        self.row_col_to_bounds = {(row_id,col_id): bounds for ((row_id,col_id),bounds) in flatten_list([
                [((row_id,col_id),((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1]))) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)]
                )}
        self.bounds = flatten_list([
                [((x_steps[col_id], y_steps[row_id]), (x_steps[col_id+1], y_steps[row_id+1])) for col_id in range(n_x_steps-1)] 
                for row_id in range(n_y_steps-1)])
        self.cell_dict = dict()
        self.clusters = dict()
        if not silent:
            print('Create grid with '+str(n_y_steps-1)+'x'+str(n_x_steps-1)+'='+str((n_y_steps-1)*(n_x_steps-1)))
        #
    #
    # add functions
    aggregate_point_data_to_cells = aggregate_point_data_to_cells
    # aggregate_point_data_to_nested_cells = aggregate_point_data_to_nested_cells
    assign_points_to_cell_regions = assign_points_to_cell_regions
    aggreagate_point_data_to_disks_vectorized = aggreagate_point_data_to_disks_vectorized
    # aggreagate_point_data_to_disks_vectorized_nested = aggreagate_point_data_to_disks_vectorized_nested
    # append a variable to pts that indicates the share of valid area float[0,1] 
    disk_cell_intersection_area = disk_cell_intersection_area

    def add_cluster_tags_to_cells(
            self,
            cells_with_cluster:_np_array,
            cluster_tag:str='employment',
    ):
        self.clusters[cluster_tag] = {
            'tag': cluster_tag,
            'cells': sorted(set(arr_to_tpls(cells_with_cluster, int)))
            }

    def merge_clusters(
            self,
            distance_thresholds:float
        ):

        for (n, (cluster_column, clusters)), distance_threshold in zip(enumerate(self.clusters.items()), distance_thresholds):
            id_to_sums = self.id_to_sums    
            row_col_to_centroid = self.row_col_to_centroid

            prime_locs = [{
                'id':i,
                'cells': [cell],
                'centroid': row_col_to_centroid[cell],
                'sum': id_to_sums[cell][n] # TODO select correct column
                } for i, cell in enumerate(clusters['cells'])]
            prime_locs_by_id = {pl['id']: pl for pl in prime_locs}
            deleted_clusters = set()
            clusters_to_delete = set([-1])
            def find_next_merge(prime_locs):
                for i, current_cluster_id in enumerate([pl['id'] for pl in prime_locs]):
                    current_cluster = prime_locs_by_id[current_cluster_id]
                    
                    # Check if current cluster has any remaining cells
                    current_cluster_cells = current_cluster['cells']
                    current_centroid = current_cluster['centroid']

                    for neighbor_cluster_id in [pl['id'] for pl in prime_locs[i+1:]]:
                        if neighbor_cluster_id == current_cluster_id or neighbor_cluster_id in clusters_to_delete:
                            continue  # Skip unclustered cells and self-comparison
                        
                        neighbor_cluster = prime_locs_by_id[neighbor_cluster_id]

                        neighbor_centroid = neighbor_cluster['centroid']

                        # Compute distance between centroids
                        # distance = geopy_distance(current_centroid, neighbor_centroid).meters
                        distance = ((current_centroid[0]-neighbor_centroid[0])**2+(current_centroid[1]-neighbor_centroid[1])**2)**.5

                        if distance < distance_threshold:
                            n_current, n_neighbor = len(current_cluster_cells), len(neighbor_cluster['cells'])
                            current_cluster['cells'] = current_cluster_cells + neighbor_cluster['cells']
                            current_cluster['sum'] += neighbor_cluster['sum']
                            current_cluster['centroid'] = (
                                (current_centroid[0]*n_current + neighbor_centroid[0]*n_neighbor)/(n_current + n_neighbor),
                                (current_centroid[1]*n_current + neighbor_centroid[1]*n_neighbor)/(n_current + n_neighbor)
                            )
                            return neighbor_cluster_id
                        #
                    #
            while len(clusters_to_delete) > 0:
                prime_locs = [c for c in prime_locs if not c['id'] in clusters_to_delete]
                prime_locs.sort(key=lambda c: (-len(c['cells']), -c['sum']))
                neighbor_cluster_id = find_next_merge(prime_locs)
                clusters_to_delete = set()
                if not neighbor_cluster_id is None:
                    clusters_to_delete.add(neighbor_cluster_id)
                    deleted_clusters.add(neighbor_cluster_id)
                    #
                #
            # assign ids starting at 1 from biggest (according to sum value) to largest cluster 
            prime_locs = [v for k,v in prime_locs_by_id.items() if not k in deleted_clusters]
            prime_locs.sort(key=lambda c: -c['sum'])
            prime_locs = [{**c, 'id':i+1} for i, c in enumerate(prime_locs)]
            clusters['prime_locs'] = prime_locs
        #
    #

    def make_cluster_convex(
            self
    ):  
        new_clusters_dict = {}
        for n, cluster_column in enumerate(self.clusters):
            clusters = self.clusters[cluster_column]
            all_clustered_cells = set()
            for cluster in clusters['prime_locs']:
                all_clustered_cells.update(cluster['cells'])
            new_prime_locs = []
            for cluster in clusters['prime_locs']:
                id_to_sums = self.id_to_sums
                row_col_to_centroid = self.row_col_to_centroid
                row_col_to_bounds = self.row_col_to_bounds
                cells = cluster['cells']
                cells_in_convex_cluster = set(cells)
                convex_hull = unary_union(
                    [[Polygon(((xmin,ymin),(xmax,ymin),(xmax,ymax),(xmin,ymax))) for (xmin,ymin),(xmax,ymax) in [self.row_col_to_bounds[cell]]][0] for cell in cells]
                ).convex_hull
                
                row_ids = sorted(set([row for row,col in cells]))
                col_ids = sorted(set([col for row,col in cells]))
                row_range = range(min(row_ids), max(row_ids)+1)
                col_range = range(min(col_ids), max(col_ids)+1)
                for r in row_range:
                    for c in col_range:
                        if not (r,c) in all_clustered_cells and convex_hull.contains(Point(row_col_to_centroid[(r,c)])):
                            cells_in_convex_cluster.add((r,c))
                            all_clustered_cells.add((r,c))
                new_prime_locs.append(
                    {**cluster,
                    'cells': sorted(cells_in_convex_cluster),
                    'sum': sum([id_to_sums[cell][n] for cell in cells_in_convex_cluster if cell in id_to_sums]),
                    'centroid': _np_array([row_col_to_centroid[cell] for cell in cells_in_convex_cluster]).sum(axis=0)/len(cells_in_convex_cluster)
                })
            new_clusters_dict[cluster_column] = {**clusters, 'prime_locs': new_prime_locs}
        
        self.clusters = new_clusters_dict
        #
    #
    def make_cluster_orthogonally_convex(
            self
        ):
        """
        ensure all cells between (=orthogononally, not diagonally) two cluster cells are also part of the cluster
        exception: a cell is part of another cluster already
        """
        id_to_sums = self.id_to_sums
        row_col_to_centroid = self.row_col_to_centroid
        for (cluster_column, clusters) in self.clusters.items():
            all_clustered_cells = set()
            for cluster in clusters['prime_locs']:
                all_clustered_cells.update(cluster['cells'])
            
            for cluster in clusters['prime_locs']:
                cells_from_other_clusters = all_clustered_cells.difference(cluster['cells'])
                n_last_it = -1
                while len(cluster['cells']) != n_last_it:
                    cells = cluster['cells']
                    cells_in_convex_cluster = set(cells)
                    row_ids = sorted(set([row for row,col in cells]))
                    col_ids = sorted(set([col for row,col in cells]))
                    # row_to_cols = {row: [c for r,c in cells if r==row] for row in row_ids}
                    # col_to_rows = {col: [r for r,c in cells if c==col] for col in col_ids}
                    # row_to_min_col = {k:min([min(v), max([
                    #     min([col for row, col in cells if row<k]),
                    #     min([col for row, col in cells if row>k])
                    # ])]) for k,v in row_to_cols.items()}
                    
                    row_range = range(min(row_ids), max(row_ids)+1)
                    col_range = range(min(col_ids), max(col_ids)+1)
                    for r in row_range:
                        cells_to_left = [col for row, col in cells if row<r]
                        cells_to_right = [col for row, col in cells if row>r]
                        cells_same_col = [col for row, col in cells if row==r]
                        max_left, min_left, max_right, min_right, max_same, min_same = None, None, None, None, None, None
                        if len(cells_to_left) > 0:
                            min_left = min(cells_to_left)
                            max_left = max(cells_to_left)
                        if len(cells_to_right) > 0:
                            min_right = min(cells_to_right)
                            max_right = max(cells_to_right)
                        if len(cells_same_col) > 0:
                            min_same = min(cells_same_col)
                            max_same = max(cells_same_col)
                        max_other = max_right if max_left is None else max_left if max_right is None else max([min_left, min_right]) 
                        min_other = min_right if min_left is None else min_left if min_right is None else min([min_left, min_right])
                        max_all = max_other if max_same is None else max_same if max_other is None else min([min_same, min_other])
                        min_all = min_other if min_same is None else min_same if min_other is None else max([min_same, min_other])
                        cells_in_convex_cluster.update([(r,c) for c in range(min_all, max_all+1)])
                    #

                    for c in col_range:
                        cells_to_left = [row for row, col in cells if col<c]
                        cells_to_right = [row for row, col in cells if col>c]
                        cells_same_col = [row for row, col in cells if col==c]
                        max_left, min_left, max_right, min_right, max_same, min_same = None, None, None, None, None, None
                        if len(cells_to_left) > 0:
                            min_left = min(cells_to_left)
                            max_left = max(cells_to_left)
                        if len(cells_to_right) > 0:
                            min_right = min(cells_to_right)
                            max_right = max(cells_to_right)
                        if len(cells_same_col) > 0:
                            min_same = min(cells_same_col)
                            max_same = max(cells_same_col)
                        # max_other = max_right if max_left is None else max_left if max_right is None or max_left < max_right else max_right 
                        # min_other = min_right if min_left is None else min_left if min_right is None or min_left > min_right else min_right
                        min_other = None if max_left is None or max_right is None else max([min_left, min_right])
                        max_other = None if max_left is None or max_right is None else min([min_left, min_right])
                        min_all = min_other if min_same is None else min_same if min_other is None else min([min_same, min_other])
                        max_all = max_other if max_same is None else max_same if max_other is None else max([min_same, min_other])
                        cells_in_convex_cluster.update([(r,c) for r in range(min_all, max_all+1)])
                    #

                    cells_in_convex_cluster.difference_update(cells_from_other_clusters)
                    cluster['cells'] = sorted(cells_in_convex_cluster)
                    n_last_it = len(cluster['cells'])
                
                cluster['aggregate_vals'] = sum([id_to_sums[cell] for cell in cells_in_convex_cluster if cell in id_to_sums])
                cluster['centroid'] = _np_array([row_col_to_centroid[cell] for cell in cells_in_convex_cluster]).sum(axis=0)/len(cells_in_convex_cluster)
            #
        #
    #
    def connect_cells_to_clusters(self):
        for (cluster_column, clusters) in self.clusters.items():
            clusters['cell_to_cluster'] = {}
            for cluster in clusters['prime_locs']:
                clusters['cell_to_cluster'].update({cell: cluster['id'] for cell in cluster['cells']})
    
    def add_geom_to_cluster(self):
        for (cluster_column, clusters) in self.clusters.items():
            for cluster in clusters['prime_locs']:
                cluster['geometry'] = unary_union(
                    [[Polygon(((xmin,ymin),(xmax,ymin),(xmax,ymax),(xmin,ymax))) for (xmin,ymin),(xmax,ymax) in [self.row_col_to_bounds[cell]]][0] for cell in cluster['cells']]
                )
                cluster['area'] = len(cluster['cells']) * self.spacing**2

    def create_clusters(
        self,
        pts:_pd_DataFrame,
        columns:list=['employment'],
        distance_thresholds=2500,
        make_convex:bool=True,
        row_name:str='id_y',
        col_name:str='id_x',
        cluster_suffix:str='_750m',
        plot_cluster_cells:dict=None,
    ):
        for column in columns:
            cells_with_cluster = (pts[[row_name, col_name]][pts[column + cluster_suffix]]).values
            self.add_cluster_tags_to_cells(
                cells_with_cluster=cells_with_cluster,
                cluster_tag=column,
            )
        
        distance_thresholds = distance_thresholds if type(distance_thresholds) in [list, _np_array] else [distance_thresholds for n in columns]
        self.merge_clusters(distance_thresholds=distance_thresholds)
        if make_convex:
            self.make_cluster_convex()
            # self.make_cluster_orthogonally_convex()
        
        self.connect_cells_to_clusters()
        self.add_geom_to_cluster()

        for column in columns:
            cluster_column = column + cluster_suffix
            cell_to_cluster = self.clusters[column]['cell_to_cluster']
            vals = _np_zeros(len(pts),int)#-1
            for i,(row,col) in enumerate(pts[[row_name, col_name]].values):
                if (row, col) in cell_to_cluster: 
                    vals[i] = cell_to_cluster[(row, col)]
            pts[cluster_column] = vals
        
        if not plot_cluster_cells is None:
            self.plot_all_clusters()

                    

    def create_sparse_grid(self):
        return
        # TODO shall this be another class? A new instane of the grid? Or create a sparse copy within itself? Or overwrite itself with an own c?
        print("Make this grid sparse")
    
    def create_nested_grid(self):
        print("Create nested grid")
    #

    
    def save_full_grid(
            self,
            filename:str="full_grid",
            file_format:str=['shp','csv'][0],
            # target_crs:str="EPSG:4326"
        ):
        for (cluster_column, clusters) in self.clusters.items():
            
            cell_to_cluster = self.clusters[cluster_column]['cell_to_cluster']
            c_ids = _np_zeros(len(self.ids),int)#-1
            sums = _np_zeros(len(self.ids),int)
            polys = []
            id_to_sums = self.id_to_sums
            centroids = _np_array(list(self.row_col_to_centroid.values()))
            transformer = Transformer.from_crs(crs_from=self.local_crs, crs_to=self.initial_crs, always_xy=True)
            centroids_x, centroids_y = transformer.transform(centroids[:,0], centroids[:,1])
            
            for (i, row_col), ((xmin,ymin),(xmax,ymax)) in zip(enumerate(self.ids), self.id_to_bounds.values()):
                if row_col in cell_to_cluster: 
                    c_ids[i] = cell_to_cluster[row_col]
                if row_col in id_to_sums: 
                    sums[i] = id_to_sums[row_col]
                polys.append(Polygon(((xmin,ymin),(xmax,ymin),(xmax,ymax),(xmin,ymax))))
            df = _gpd_GeoDataFrame({
                'centroid_x': centroids_x,
                'centroid_y': centroids_y,
                'cluster_id': c_ids,
                'sum': sums,},
                geometry=polys,
                crs=self.local_crs
                )
            df.to_crs(self.initial_crs, inplace=True)
            # save
            filename = filename + (("_"+cluster_column) if len(self.clusters) > 1 else '') + '.'+file_format
            if file_format == 'shp':
                df.to_file(filename, driver="ESRI Shapefile", index=False)
            elif file_format == 'csv':
                df.to_csv(filename, index=False)
            else:
                raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        #
    
    def save_sparse_grid(
            self,
            filename:str="sparse_grid",
            file_format:str=['shp','csv'][0],
            # target_crs:str="EPSG:4326"
        ):
        """
        saving all grid cell filled with
        """
        for (cluster_column, clusters) in self.clusters.items():
            
            cell_to_cluster = self.clusters[cluster_column]['cell_to_cluster']
            id_to_sums = self.id_to_sums
            polys = []
            c_ids = _np_zeros(len(id_to_sums),int)
            sums = _np_zeros(len(id_to_sums),float)
            centroids = _np_array(list(self.row_col_to_centroid.values()))
            transformer = Transformer.from_crs(crs_from=self.local_crs, crs_to=self.initial_crs, always_xy=True)
            centroids_x_full, centroids_y_full = transformer.transform(centroids[:,0], centroids[:,1])
            centroids_x = _np_zeros(len(id_to_sums),float)
            centroids_y = _np_zeros(len(id_to_sums),float)
            i = 0
            for row_col, ((xmin,ymin),(xmax,ymax)), c_x, c_y in zip(self.ids, self.id_to_bounds.values(), centroids_x_full, centroids_y_full):
                if row_col in id_to_sums: 
                    if row_col in cell_to_cluster: 
                        c_ids[i] = cell_to_cluster[row_col]
                    sums[i] = id_to_sums[row_col]
                    centroids_x[i] = c_x 
                    centroids_y[i] = c_y 
                    polys.append(Polygon(((xmin,ymin),(xmax,ymin),(xmax,ymax),(xmin,ymax))))
                    i += 1
                #
            #
            df = _gpd_GeoDataFrame({
                'centroid_x': centroids_x,
                'centroid_y': centroids_y,
                'cluster_id': c_ids,
                'sum': sums,},
                geometry=polys,
                crs=self.local_crs
                )
            df.to_crs(self.initial_crs, inplace=True)
            # save
            filename = filename + (("_"+cluster_column) if len(self.clusters) > 1 else '') + '.'+file_format
            if file_format == 'shp':
                df.to_file(filename, driver="ESRI Shapefile", index=False)
            elif file_format == 'csv':
                df.to_csv(filename, index=False)
            else:
                raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        #
    
    def save_grid_clusters(
            self,
            filename:str="grid_clusters",
            file_format:str=['shp','csv'][0],
            # target_crs:str="EPSG:4326"
        ):
        for (cluster_column, clusters) in self.clusters.items():
            df = _gpd_GeoDataFrame({
                'centroid_x': [pl['centroid'][0] for pl in clusters['prime_locs']],
                'centroid_y': [pl['centroid'][1] for pl in clusters['prime_locs']],
                'cluster_id': [pl['id'] for pl in clusters['prime_locs']],
                'sum': [pl['sum'] for pl in clusters['prime_locs']],
                "n_cells": [len(pl['cells']) for pl in clusters['prime_locs']],
                'area': [pl['area'] for pl in clusters['prime_locs']],
                },
                geometry = [pl['geometry'] for pl in clusters['prime_locs']],
                crs=self.local_crs)
            transformer = Transformer.from_crs(crs_from=self.local_crs, crs_to=self.initial_crs, always_xy=True)
            df['centroid_x'], df['centroid_y'] = transformer.transform(df['centroid_x'], df['centroid_y'])
            df.to_crs(self.initial_crs, inplace=True)
            # save
            filename = filename + (("_"+cluster_column) if len(self.clusters) > 1 else '') + '.'+file_format
            if file_format == 'shp':
                df.to_file(filename, driver="ESRI Shapefile", index=False)
            elif file_format == 'csv':
                df.to_csv(filename, index=False)
            else:
                raise ValueError('Unknown file_format:',file_format,'Choose on out of shp, csv')
        #
    #

    def plot_cell_sums(
            self, fig=None, ax=None, filename:str=''):
        if ax is None:
            fig, ax = plt.subplots(nrows=len(self.clusters), figsize=(10,10))
        
        id_to_sums = self.id_to_sums
        imshow_kwargs = {
            'xmin':self.x_steps.min(),
            'ymin':self.y_steps.min(),
            'xmax':self.x_steps.max(),
            'ymax':self.y_steps.max(),
        }
        
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]
        cmap = _plt_get_cmap('Reds')
        cmap.set_under('#ccc')
        max_sum = max(list(id_to_sums.values()))
        X = _np_array([[id_to_sums[(row,col)][0] if ((row,col)) in id_to_sums else 0 for col in  self.col_ids] for row in self.row_ids])
        ux = _np_unique(X)
        minX = min(ux[ux!=0])
        print("min X", min(X), max(X))
        # p = ax.imshow(X=X, interpolation='none', cmap=cmap, vmin=1e-5,vmax=max_sum, extent=extent)
        p = ax.pcolormesh(X, cmap=cmap, vmin=minX/2,vmax=max_sum)
        cb = plt.colorbar(p)
        ax.set_xlabel('x/lon') 
        ax.set_ylabel('y/lat') 
        ax.title.set_text('Sums in grid cells')
        if filename:
            fig.savefig(filename, dpi=300, bbox_inches="tight")
    


    def plot_grid_ids(self, fig=None, ax=None, filename:str='',):
        if ax is None:
            fig, ax = plt.subplots(nrows=3, figsize=(15,25))
        imshow_kwargs = {
            'xmin':self.x_steps.min(),
            'ymin':self.y_steps.min(),
            'xmax':self.x_steps.max(),
            'ymax':self.y_steps.max(),
        }
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]
        X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.x_steps[:-1]] for y in self.y_steps[:-1]])
        # ax.flat[0].imshow(X=X, interpolation='none', extent=extent)
        # ax.flat[0].pcolormesh([self.x_steps, self.y_steps], X)
        ax.flat[0].pcolormesh(X, edgecolor="black", linewidth=1/max([self.n_x_steps, self.n_y_steps])/1.35)
        # ax.flat[0].set_aspect(2)
        colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
        cb = plt.colorbar(**colorbar_kwargs[2], ax=ax.flat[0])
        cb.ax.set_xlabel("diagonal")
        cb = plt.colorbar(**colorbar_kwargs[0], ax=ax.flat[0])
        cb.ax.set_xlabel("x/lon")
        cb = plt.colorbar(**colorbar_kwargs[1], ax=ax.flat[0])
        cb.ax.set_xlabel("y/lat") 
        ax.flat[0].set_xlabel('x/lon') 
        ax.flat[0].set_ylabel('y/lat') 
        ax.flat[0].title.set_text("Grid lat / lon coordinates")
        # ax.flat[0].set_xticks(self.x_steps, minor=True)
        # ax.flat[0].set_yticks(self.y_steps, minor=True)
        # ax.flat[0].grid(which='minor', color='w', linestyle='-', linewidth=0.002)

        imshow_kwargs = {
            'xmin':self.col_ids.min(),
            'ymin':self.row_ids.min(),
            'xmax':self.col_ids.max(),
            'ymax':self.row_ids.max(),
        }
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]

        X = _np_array([[map_2D_to_rgb(x,y, **imshow_kwargs) for x in  self.col_ids] for y in self.row_ids])
        ax.flat[1].imshow(X=X, interpolation='none', extent=extent)
        # ax.flat[1].set_aspect(2)
        colorbar_kwargs = get_2D_rgb_colobar_kwargs(**imshow_kwargs)
        cb = plt.colorbar(**colorbar_kwargs[2], ax=ax.flat[1])
        cb.ax.set_xlabel("diagonal")
        cb = plt.colorbar(**colorbar_kwargs[0], ax=ax.flat[1])
        cb.ax.set_xlabel("col nr")
        cb = plt.colorbar(**colorbar_kwargs[1], ax=ax.flat[1])
        cb.ax.set_xlabel("row nr") 
        ax.flat[1].set_xlabel('row nr') 
        ax.flat[1].set_ylabel('col nr') 
        ax.flat[1].title.set_text("Grid row / col indices")
        # ax.flat[1].set_xticks(self.col_ids, minor=True)
        # ax.flat[1].set_yticks(self.row_ids, minor=True)
        # ax.flat[1].grid(which='minor', color='black', linestyle='-', linewidth=0.003)
        
        X = _np_array([[len(self.id_to_pt_ids[(row_id, col_id)]) if (row_id, col_id) in self.id_to_pt_ids else 0 for col_id in self.col_ids] for row_id in self.row_ids])
        p = ax.flat[2].pcolormesh(X, cmap='Reds')
        ax.flat[2].set_xlabel('row nr') 
        ax.flat[2].set_ylabel('col nr') 
        plt.colorbar(p)
        if filename:
            fig.savefig(filename, dpi=300, bbox_inches="tight")
    
    def plot_clusters(self, cluster_column:str, fig=None, ax=None, filename:str='',):
        if ax is None:
            fig, ax = plt.subplots(nrows=len(self.clusters), figsize=(15,25))
        
        clusters = self.clusters[cluster_column]
        cell_to_cluster = clusters['cell_to_cluster']
        max_cluster_id = max(list(clusters['cell_to_cluster'].values()))
        imshow_kwargs = {
            'xmin':self.x_steps.min(),
            'ymin':self.y_steps.min(),
            'xmax':self.x_steps.max(),
            'ymax':self.y_steps.max(),
        }
        
        extent=[imshow_kwargs['xmin'],imshow_kwargs['xmax'],imshow_kwargs['ymax'],imshow_kwargs['ymin']]
        cmap = _plt_get_cmap('Reds')
        cmap.set_under('#ccc')
        X = _np_array([[cell_to_cluster[(row,col)]/max_cluster_id if (row,col) in cell_to_cluster else 0 for col in  self.col_ids] for row in self.row_ids])
        p = ax.imshow(X=X, interpolation='none', cmap=cmap, vmin=.5/max_cluster_id,vmax=1.0, extent=extent)
        # p = ax.pcolormesh(X, cmap=cmap, edgecolor="black", linewidth=1/max([self.n_x_steps, self.n_y_steps])/1.35)
        # cb = plt.colorbar(p)
        for cluster in clusters['prime_locs']:
            ax.annotate(cluster['id'], xy=cluster['centroid'], fontsize=10)
        ax.set_xlabel('x/lon') 
        ax.set_ylabel('y/lat') 
        ax.title.set_text(str(len(clusters['prime_locs']))+' clusters for '+str(cluster_column))
    
    def plot_all_clusters(self, fig=None, axs=None, filename:str=''):
        
        if axs is None:
            fig, axs = plt.subplots(nrows=len(self.clusters), figsize=(10,10*len(self.clusters)))

        for i, cluster_column in enumerate(self.clusters):
            ax = axs if not hasattr(axs, 'flat') else axs.flat[i]
            self.plot_clusters(cluster_column=cluster_column, ax=ax)
        if filename:
            fig.savefig(filename, dpi=300, bbox_inches="tight")
#

class ExludedArea:
    def __init__(self,excluded_area_geometry_or_list, grid:Grid):
        # recursively split exluded area geometry along grid 
        # then sort it into grid cell
        
        pass
#





import numpy as np

i8 = np.int64()
i4 = np.int32()
u8 = np.uint64()
b_ = np.bool()
i = int()

f8 = np.float64()

b_ >> f8  # E: No overload variant
i8 << f8  # E: No overload variant
i | f8  # E: Unsupported operand types
i8 ^ f8  # E: No overload variant
u8 & f8  # E: No overload variant
~f8  # E: Unsupported operand type
# TODO: Certain mixes like i4 << u8 go to float and thus should fail

# mypys' error message for `NoReturn` is unfortunately pretty bad
# TODO: Re-enable this once we add support for numerical precision for `number`s
# a = u8 | 0  # E: Need type annotation



from collections.abc import Sequence
from numpy._typing import _NestedSequence

a: Sequence[float]
b: list[complex]
c: tuple[str, ...]
d: int
e: str

def func(a: _NestedSequence[int]) -> None:
    ...

reveal_type(func(a))  # E: incompatible type
reveal_type(func(b))  # E: incompatible type
reveal_type(func(c))  # E: incompatible type
reveal_type(func(d))  # E: incompatible type
reveal_type(func(e))  # E: incompatible type



import aabpl

# def test_aabpl():
#     """
#     test invert_quality_of_life with testdata and default parameters
#     """
#     try:
#         aabpl.invert_quality_of_life(df=aabpl.testdata)
#         success=True
#     except:
#         success=False
#     assert success
    



import math as math

from numpy._pytesttester import PytestTester

from numpy import (
    ndenumerate as ndenumerate,
    ndindex as ndindex,
)

from numpy.version import version

from numpy.lib import (
    format as format,
    mixins as mixins,
    scimath as scimath,
    stride_tricks as stride_tricks,
    npyio as npyio,
    array_utils as array_utils,
)

from numpy.lib._version import (
    NumpyVersion as NumpyVersion,
)

from numpy.lib._arrayterator_impl import (
    Arrayterator as Arrayterator,
)

from numpy._core.multiarray import (
    add_docstring as add_docstring,
    tracemalloc_domain as tracemalloc_domain,
)

from numpy._core.function_base import (
    add_newdoc as add_newdoc,
)

__all__: list[str]
test: PytestTester

__version__ = version
